import torch
import torch.nn as nn

# CrossEntropyLoss with additional cross weight loss for two targets
from NewsSentiment.SentimentClasses import SentimentClasses


class CrossEntropyLoss_CrossWeight(nn.Module):
    def __init__(self, device, ignore_index, weight=None, crossloss_weight=0.2):
        super(CrossEntropyLoss_CrossWeight, self).__init__()
        self.device = device
        self.class_loss = nn.CrossEntropyLoss(weight=weight, ignore_index=ignore_index,)
        # cosine similarity by default
        self.cross_weight_loss = nn.CosineSimilarity(dim=2)
        self.crossloss_weight = crossloss_weight
        self.crossentropyloss_weight = 1.0 - crossloss_weight
        self.ignore_index = ignore_index

    def forward(
        self,
        predictions: torch.Tensor,
        true_labels: torch.Tensor,
        cross_weight: torch.Tensor,
        target_mask: torch.Tensor,
    ):
        """
        Calculate k-dimensional cross entropy loss by comparing predictions and
        true_labels. Additionally, calculates a cross_weight loss that maximizes the
        differences between the weights compared between both targets.

        :param predictions: shape: (batch, polarities e.g. 3, targets ie 2)
        :param true_labels: shape: (batch, targets)
        :param cross_weight:shape: (batch, targets, seqlen, bertdim)
        :return:
        prediction size earlier was batch polarities
        """
        batch_size, num_classes, num_targets = predictions.size()
        bert_dim = cross_weight.shape[3]

        # calculate regular cross entropy loss
        classification_loss = self.class_loss(predictions, true_labels)

        # calculate cross-weight loss
        # we must not calc the cross weight loss for batch item of two targets,
        # where 1 or 2 is a fill up target, i.e., the true value is
        # SentimentClasses.FILLUP_POLARITY_VALUE
        # get all batch items where the first target is fill up
        is_ignored_a = true_labels[:, 0]
        is_ignored_a = is_ignored_a == SentimentClasses.FILLUP_POLARITY_VALUE
        # get all batch items where the second is
        is_ignored_b = true_labels[:, 1]
        is_ignored_b = is_ignored_b == SentimentClasses.FILLUP_POLARITY_VALUE
        # logical or both lists to one
        is_ignored_batch_item = is_ignored_a | is_ignored_b
        # select only those batch items where no target
        is_not_ignored_batch_item = ~is_ignored_batch_item
        target_mask = target_mask[is_not_ignored_batch_item, :, :]
        count_non_ignored_batch_items = target_mask.shape[0]

        # if we have identical targets (=target masks) set cross weight loss to 0
        target_mask_a = target_mask[:, 0, :]
        target_mask_b = target_mask[:, 1, :]
        diff_target_mask = target_mask_a - target_mask_b
        diff_target_mask = diff_target_mask.sum(dim=1)
        # shape: batch
        # diff target mask, will be 0 if the two targets in one batched item are
        # identical, and 1 if they are different
        is_different_target_per_two_batch_items = diff_target_mask != 0

        # only selecting different targets effectively zeros out values of identical
        # targets
        cross_weight = cross_weight[is_different_target_per_two_batch_items]

        # if there is not at least a single batch item with different targets, shape[0]
        # will be 0. we use this to test for this condition and if all targets over all
        # batch items are identical, we skip the cross weight part
        count_different_targets = cross_weight.shape[0]
        if count_different_targets == 0:
            cross_weight_loss = 0
        else:
            assert count_different_targets >= 1
            seq_len = cross_weight.shape[2]
            weight_a = cross_weight[:, 0, :, :]
            # weight_b = cross_weight[:, 1, :, :]
            weight_b = weight_a

            # we add the negative sign, since we actually want to maximize the distance
            # between both vectors
            cross_weight_similarity = self.cross_weight_loss(weight_a, weight_b)
            # cross_weight_similarity will be -1 for absolutely dissimilar values
            # 0 for unrelated and +1 for identical values
            # normalize between 0 and 1
            cross_weight_similarity = (cross_weight_similarity + 1) / 2
            # 0 = dissimilar, 1 = identical
            # mean over seq len
            cross_weight_loss = cross_weight_similarity.mean(dim=1)
            # at this point we have for each batch item its loss (0 if dissimilar
            # targets, 1 if identical target in the batch item)
            # finally, compute the single loss: sum
            cross_weight_loss = cross_weight_loss.sum()
            # normalize (divide by number of batch items; note that this can be different
            # from the number of different targets in the batch)
            cross_weight_loss = cross_weight_loss / count_non_ignored_batch_items

        # total_loss
        total_loss = (
            self.crossentropyloss_weight * classification_loss
            + self.crossloss_weight * cross_weight_loss
        )

        return total_loss



# adapted from https://github.com/yangheng95/LC-ABSA/blob/c945a94e0f86116c5578245aa9ad36c46c7b9c4a/models/lc_apc/lcf_bert.py
# according to
import copy
from argparse import Namespace
from typing import Dict

import numpy as np
import torch
import torch.nn as nn
from transformers.modeling_bert import BertPooler, BertSelfAttention

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.layers.attention import FXBertSelfAttention
from NewsSentiment.models.FXBaseModel import FXBaseModel


class GlobalContext(nn.Module):
    def __init__(self, global_context_seqs_per_doc):
        super(GlobalContext, self).__init__()
        self.global_context_seqs_per_doc = global_context_seqs_per_doc

    def forward(self, inputs):
        pass


class SelfAttention(nn.Module):
    def __init__(self, config, opt):
        super(SelfAttention, self).__init__()
        self.opt = opt
        self.config = config
        self.SA = FXBertSelfAttention(
            hidden_size=config.hidden_size,
            num_attention_heads=config.num_attention_heads,
            attention_probs_dropout_prob=0.1,
        )
        self.tanh = torch.nn.Tanh()

    def forward(self, inputs):
        zero_tensor = torch.tensor(
            np.zeros((inputs.size(0), 1, 1, self.opt.max_seq_len), dtype=np.float32),
            dtype=torch.float32,
        ).to(self.opt.device)
        SA_out = self.SA(inputs, zero_tensor)
        return self.tanh(SA_out[0])


class LCF_BERT2Dual(FXBaseModel):
    """
    While lcf.py:LCF_BERT is the implementation as implemented in PyTorch-ABSA repository, this implementation here
    (LCF_BERT2Dual) is following the implementation as in the author's repository, which according to
    https://github.com/yangheng95/LC-ABSA/issues/10#issuecomment-670301603 has seen some more improvements compared to
    the version from PyTorch-ABSA
    """

    @staticmethod
    def get_language_models():
        return (get_default_lm(),)

    @staticmethod
    def get_input_field_ids():
        return [
            (get_default_lm(), FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS),
            (
                get_default_lm(),
                FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_SEGMENT_IDS,
            ),
            (get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (get_default_lm(), FIELD_TARGET_IDS_WITH_SPECIAL_TOKENS),
            (get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(LCF_BERT2Dual, self).__init__()

        bert = transformer_models[get_default_lm()]

        self.bert4global = bert
        # note that we use a second bert here, which should slightly improve performance
        # cf. https://github.com/yangheng95/LC-ABSA/#tips
        # self.bert4local = copy.deepcopy(bert)
        # we can't do this on scc because even for batch size = only 16 we run out of
        # memory. because of that, we use the same bert for both local and global
        # (just as in lcf.py)
        self.bert4local = bert
        self.opt = opt
        self.dropout = nn.Dropout(self.opt.dropout)
        self.bert_SA = SelfAttention(bert.config, self.opt)
        self.linear2 = nn.Linear(bert.config.hidden_size * 2, bert.config.hidden_size)
        # self.linear3 = nn.Linear(bert.config.hidden_size * 3, bert.config.hidden_size)
        self.bert_pooler = BertPooler(bert.config)
        self.dense = nn.Linear(bert.config.hidden_size, self.opt.polarities_dim)

    def feature_dynamic_mask(self, text_local_indices, aspect_indices):
        texts = text_local_indices.cpu().numpy()
        asps = aspect_indices.cpu().numpy()
        mask_len = self.opt.SRD
        masked_text_raw_indices = np.ones(
            (
                text_local_indices.size(0),
                self.opt.max_seq_len,
                self.bert4local.config.hidden_size,
            ),
            dtype=np.float32,
        )
        for text_i, asp_i in zip(range(len(texts)), range(len(asps))):
            asp_len = np.count_nonzero(asps[asp_i]) - 2
            try:
                asp_begin = np.argwhere(texts[text_i] == asps[asp_i][1])[0][0]
            except:
                continue
            if asp_begin >= mask_len:
                mask_begin = asp_begin - mask_len
            else:
                mask_begin = 0
            for i in range(mask_begin):
                masked_text_raw_indices[text_i][i] = np.zeros(
                    (self.bert4local.config.hidden_size), dtype=np.float
                )
            for j in range(asp_begin + asp_len + mask_len, self.opt.max_seq_len):
                masked_text_raw_indices[text_i][j] = np.zeros(
                    (self.bert4local.config.hidden_size), dtype=np.float
                )
        masked_text_raw_indices = torch.from_numpy(masked_text_raw_indices)
        return masked_text_raw_indices.to(self.opt.device)

    def feature_dynamic_weighted(self, text_local_indices, aspect_indices):
        texts = text_local_indices.cpu().numpy()
        asps = aspect_indices.cpu().numpy()
        masked_text_raw_indices = np.ones(
            (
                text_local_indices.size(0),
                self.opt.max_seq_len,
                self.bert4local.config.hidden_size,
            ),
            dtype=np.float32,
        )
        for text_i, asp_i in zip(range(len(texts)), range(len(asps))):
            asp_len = np.count_nonzero(asps[asp_i]) - 2
            try:
                asp_begin = np.argwhere(texts[text_i] == asps[asp_i][1])[0][0]
                asp_avg_index = (asp_begin * 2 + asp_len) / 2
            except:
                continue
            distances = np.zeros(np.count_nonzero(texts[text_i]), dtype=np.float32)
            for i in range(1, np.count_nonzero(texts[text_i]) - 1):
                if abs(i - asp_avg_index) + asp_len / 2 > self.opt.SRD:
                    distances[i] = 1 - (
                        abs(i - asp_avg_index) + asp_len / 2 - self.opt.SRD
                    ) / np.count_nonzero(texts[text_i])
                else:
                    distances[i] = 1
            for i in range(len(distances)):
                masked_text_raw_indices[text_i][i] = (
                    masked_text_raw_indices[text_i][i] * distances[i]
                )
        masked_text_raw_indices = torch.from_numpy(masked_text_raw_indices)
        return masked_text_raw_indices.to(self.opt.device)

    def forward(self, inputs):
        text_target_bert_indices = FXDataset.get_input_by_params(
            inputs, get_default_lm(), FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS,
        )
        text_target_bert_segments_ids = FXDataset.get_input_by_params(
            inputs,
            get_default_lm(),
            FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_SEGMENT_IDS,
        )
        text_local_indices = FXDataset.get_input_by_params(
            inputs, get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        aspect_indices = FXDataset.get_input_by_params(
            inputs, get_default_lm(), FIELD_TARGET_IDS_WITH_SPECIAL_TOKENS
        )

        # bert
        global_context_features = self.invoke_language_model(
            self.bert4global,
            input_ids=text_target_bert_indices,
            token_type_ids=text_target_bert_segments_ids,
        )
        local_context_features = self.invoke_language_model(
            self.bert4local, text_local_indices
        )

        # mask
        if self.opt.local_context_focus == "cdm":
            lcf_matrix = self.feature_dynamic_mask(text_local_indices, aspect_indices)
        elif self.opt.local_context_focus == "cdw":
            lcf_matrix = self.feature_dynamic_weighted(
                text_local_indices, aspect_indices
            )

        # LCF layer
        lcf_features = torch.mul(local_context_features, lcf_matrix)
        lcf_features = self.bert_SA(lcf_features)

        cat_features = torch.cat((lcf_features, global_context_features), dim=-1)
        cat_features = self.linear2(cat_features)
        cat_features = self.dropout(cat_features)

        pooled_out = self.bert_pooler(cat_features)
        dense_out = self.dense(pooled_out)

        return dense_out



# adapted from https://github.com/yangheng95/LCF-ABSA
from argparse import Namespace
from typing import Dict

import numpy as np
import torch
import torch.nn as nn
from transformers.modeling_bert import BertPooler, BertSelfAttention

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.models.FXBaseModel import FXBaseModel


class GlobalContext(nn.Module):
    def __init__(self, global_context_seqs_per_doc):
        super(GlobalContext, self).__init__()
        self.global_context_seqs_per_doc = global_context_seqs_per_doc

    def forward(self, inputs):
        pass


class SelfAttention(nn.Module):
    def __init__(self, config, opt):
        super(SelfAttention, self).__init__()
        self.opt = opt
        self.config = config
        self.SA = BertSelfAttention(config)
        self.tanh = torch.nn.Tanh()

    def forward(self, inputs):
        zero_tensor = torch.tensor(
            np.zeros((inputs.size(0), 1, 1, self.opt.max_seq_len), dtype=np.float32),
            dtype=torch.float32,
        ).to(self.opt.device)
        SA_out = self.SA(inputs, zero_tensor)
        return self.tanh(SA_out[0])


class LCF_BERT(FXBaseModel):
    @staticmethod
    def get_language_models():
        return (BERT_BASE_UNCASED,)

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS),
            (
                BERT_BASE_UNCASED,
                FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_SEGMENT_IDS,
            ),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TARGET_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(LCF_BERT, self).__init__()

        bert = transformer_models[BERT_BASE_UNCASED]
        self.bert_spc = bert
        self.opt = opt

        # self.bert_local = copy.deepcopy(bert)  # Uncomment the line to use dual Bert
        self.bert_local = (
            bert  # Default to use single Bert and reduce memory requirements
        )
        self.dropout = nn.Dropout(self.opt.dropout)
        # while the paper describes 3 self attentions, the original implementation by the authors
        # uses only one. we stick with the original implementation.
        # answer by the author: the version found PyTorch-ABSA repository and below is better than what was
        # described in the paper (cf. https://github.com/yangheng95/LC-ABSA/issues/10#issuecomment-670301603)
        # self.bert_local_SA = SelfAttention(bert.config, self.opt)
        # self.bert_global_SA = SelfAttention(bert.config, self.opt)
        self.linear_double = nn.Linear(
            bert.config.hidden_size * 2, bert.config.hidden_size
        )
        self.bert_SA = SelfAttention(bert.config, self.opt)
        self.linear_single = nn.Linear(bert.config.hidden_size, bert.config.hidden_size)
        self.bert_pooler = BertPooler(bert.config)

        self.dense = nn.Linear(bert.config.hidden_size, self.opt.polarities_dim)

    def feature_dynamic_mask(self, text_local_indices, aspect_indices):
        texts = text_local_indices.cpu().numpy()
        asps = aspect_indices.cpu().numpy()
        mask_len = self.opt.SRD
        masked_text_raw_indices = np.ones(
            (
                text_local_indices.size(0),
                self.opt.max_seq_len,
                self.bert_local.config.hidden_size,
            ),
            dtype=np.float32,
        )
        for text_i, asp_i in zip(range(len(texts)), range(len(asps))):
            asp_len = np.count_nonzero(asps[asp_i]) - 2
            try:
                asp_begin = np.argwhere(texts[text_i] == asps[asp_i][1])[0][0]
            except:
                continue
            if asp_begin >= mask_len:
                mask_begin = asp_begin - mask_len
            else:
                mask_begin = 0
            for i in range(mask_begin):
                masked_text_raw_indices[text_i][i] = np.zeros(
                    (self.bert_local.config.hidden_size), dtype=np.float
                )
            for j in range(asp_begin + asp_len + mask_len, self.opt.max_seq_len):
                masked_text_raw_indices[text_i][j] = np.zeros(
                    (self.bert_local.config.hidden_size), dtype=np.float
                )
        masked_text_raw_indices = torch.from_numpy(masked_text_raw_indices)
        return masked_text_raw_indices.to(self.opt.device)

    def feature_dynamic_weighted(self, text_local_indices, aspect_indices):
        texts = text_local_indices.cpu().numpy()
        asps = aspect_indices.cpu().numpy()
        masked_text_raw_indices = np.ones(
            (
                text_local_indices.size(0),
                self.opt.max_seq_len,
                self.bert_local.config.hidden_size,
            ),
            dtype=np.float32,
        )
        for text_i, asp_i in zip(range(len(texts)), range(len(asps))):
            asp_len = np.count_nonzero(asps[asp_i]) - 2
            try:
                asp_begin = np.argwhere(texts[text_i] == asps[asp_i][1])[0][0]
                asp_avg_index = (asp_begin * 2 + asp_len) / 2
            except:
                continue
            distances = np.zeros(np.count_nonzero(texts[text_i]), dtype=np.float32)
            for i in range(1, np.count_nonzero(texts[text_i]) - 1):
                if abs(i - asp_avg_index) + asp_len / 2 > self.opt.SRD:
                    distances[i] = 1 - (
                        abs(i - asp_avg_index) + asp_len / 2 - self.opt.SRD
                    ) / np.count_nonzero(texts[text_i])
                else:
                    distances[i] = 1
            for i in range(len(distances)):
                masked_text_raw_indices[text_i][i] = (
                    masked_text_raw_indices[text_i][i] * distances[i]
                )
        masked_text_raw_indices = torch.from_numpy(masked_text_raw_indices)
        return masked_text_raw_indices.to(self.opt.device)

    def forward(self, inputs):
        text_target_bert_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS,
        )

        text_target_bert_segments_ids = FXDataset.get_input_by_params(
            inputs,
            BERT_BASE_UNCASED,
            FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_SEGMENT_IDS,
        )
        text_local_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        aspect_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TARGET_IDS_WITH_SPECIAL_TOKENS
        )

        # apply bert and dropout
        bert_spc_out, _, _ = self.bert_spc(
            text_target_bert_indices, text_target_bert_segments_ids
        )
        bert_spc_out = self.dropout(bert_spc_out)

        bert_local_out, _, _ = self.bert_local(text_local_indices)
        bert_local_out = self.dropout(bert_local_out)

        if self.opt.local_context_focus == "cdm":
            masked_local_text_vec = self.feature_dynamic_mask(
                text_local_indices, aspect_indices
            )
            bert_local_out = torch.mul(bert_local_out, masked_local_text_vec)
        elif self.opt.local_context_focus == "cdw":
            weighted_text_local_features = self.feature_dynamic_weighted(
                text_local_indices, aspect_indices
            )
            bert_local_out = torch.mul(bert_local_out, weighted_text_local_features)

        # attention
        # bert_local_out = self.bert_local_SA(bert_local_out)
        # bert_spc_out = self.bert_global_SA(bert_spc_out)

        # cat
        out_cat = torch.cat((bert_local_out, bert_spc_out), dim=-1)

        # "interactive learning layer"
        mean_pool = self.linear_double(out_cat)
        self_attention_out = self.bert_SA(mean_pool)
        pooled_out = self.bert_pooler(self_attention_out)

        dense_out = self.dense(pooled_out)
        return dense_out



"""
Definitions:
setup: represents a fixed, unique combination of for instance:
* model, e.g., BERT, DISTILBERT
* hyper-parameters, such as learning rate loss function, weights of classes for loss function
* others, such as softmax,
* input style, e.g., input style: QA style, AEN, ...

Procedure:
Given a single annotated dataset named alldat (also called both.jsonl, since it consists of both splits of the devtest
set):
* splits alldat into testdat and remainderdat

Given a set of experiment setup descriptions
1) Iterate setup descriptions, for each:
a) create 10-fold CV to well evaluate the performance of that particular setup
b) return model from best epoch (or from best fold???!?)
2) retrievel all best performing models and test them on testdat
3) return model that performs best on testdat
"""
from threading import Lock
import argparse
import multiprocessing
import os
import subprocess
import time
from collections import Counter, defaultdict
from datetime import datetime
from itertools import product

from jsonlines import jsonlines
from tabulate import tabulate
from tqdm import tqdm

from NewsSentiment.combinations_ensemble import combinations_ensemble
from NewsSentiment.combinations_final_lcf2 import combinations_final_lcf2
from NewsSentiment.combinations_final_nostance import combinations_final_nostance
from NewsSentiment.combinations_final_spc import combinations_final_spc
from NewsSentiment.combinations_final_stance0 import combinations_final_stance0
from NewsSentiment.combinations_final_stance1 import combinations_final_stance1
from NewsSentiment.combinations_final_tdbert import combinations_final_tdbert
from NewsSentiment.combinations_top import combinations_top
from NewsSentiment.DatasetPreparer import DatasetPreparer
from NewsSentiment.combinations_default import combinations_default_0
from NewsSentiment.combinations_top4otherdomains import combinations_top4otherdomains
from NewsSentiment.combinations_top_coref16 import combinations_top_coref16
from NewsSentiment.combinations_top_coref8 import combinations_top_coref8
from NewsSentiment.combinations_top_coref8_stanceno import combinations_top_coref8_stanceno
from NewsSentiment.combinations_top_finetuned import combinations_top_finetuned
from NewsSentiment.diskdict import DiskDict
from NewsSentiment.fxlogger import get_logger
from NewsSentiment.train import str2bool

completed_tasks = None  # will be DiskDict later
completed_tasks_in_this_run_count = 0
lock = Lock()


def start_worker(
    experiment_id,
    experiment_named_id,
    named_combination,
    cmd,
    human_cmd,
    experiment_path,
    running_processes,
):
    if "--device" in cmd:
        device_index_in_list = cmd.index("--device")
        device = cmd[device_index_in_list + 1]
    else:
        device = -1

    running_processes[experiment_id] = (True, device)
    logger = get_logger()

    logger.debug("starting single setup: {}".format(human_cmd))
    with open(os.path.join(experiment_path, "stdlog.out"), "w") as file_stdout, open(
        os.path.join(experiment_path, "stdlog.err"), "w"
    ) as file_stderr:
        completed_process = subprocess.run(cmd, stdout=file_stdout, stderr=file_stderr)

    experiment_details = get_experiment_result_detailed(experiment_path)

    running_processes[experiment_id] = (False, device)
    return {
        **named_combination,
        **{"rc": completed_process.returncode, "experiment_id": experiment_id},
        "details": experiment_details,
        "experiment_named_id": experiment_named_id,
    }


def on_task_done(x):
    with lock:
        # result_list is modified only by the main process, not the pool workers.
        completed_tasks[x["experiment_named_id"]] = x
        completed_tasks.sync_to_disk()
        global completed_tasks_in_this_run_count
        completed_tasks_in_this_run_count += 1


def on_task_error(x):
    with lock:
        # result_list is modified only by the main process, not the pool workers.
        completed_tasks[x["experiment_named_id"]] = x
        completed_tasks.sync_to_disk()
        global completed_tasks_in_this_run_count
        completed_tasks_in_this_run_count += 1


def get_experiment_result_detailed(experiment_path):
    experiment_results_path = os.path.join(experiment_path, "experiment_results.jsonl")
    try:
        with jsonlines.open(experiment_results_path, "r") as reader:
            lines = []
            for line in reader:
                lines.append(line)
            assert len(lines) == 1
        return lines[0]
    except FileNotFoundError:
        return None


class SetupController:
    def __init__(self, options):
        self.logger = get_logger()
        self.opt = options

        if self.opt.cuda_devices:
            # to run on SCC
            if self.opt.cuda_devices == "SGE_GPU":
                self.cuda_devices = os.environ.get("SGE_GPU")
            else:
                self.cuda_devices = self.opt.cuda_devices

            if self.cuda_devices:
                self.logger.info("cuda devices:" + self.cuda_devices)
                self.cuda_devices = self.cuda_devices.split(",")
                self.logger.info(
                    f"was assigned {len(self.cuda_devices)} cuda devices: "
                    f"{self.cuda_devices}"
                )
                if self.opt.num_workers < 0:
                    self.logger.info(
                        f"num_workers < 0: using cuda device count. setting "
                        f"num_workers={len(self.cuda_devices)}"
                    )
                    self.opt.num_workers = len(self.cuda_devices)

        else:
            # do not use CUDA
            self.cuda_devices = None

        self.snem = "f1_macro"
        self.experiment_base_path = self.opt.experiments_path

        args_names_ordered = [
            "own_model_name",
            "optimizer",
            "initializer",
            "learning_rate",
            "batch_size",
            "balancing",
            "num_epoch",
            "loss",
            "eval_only_after_last_epoch",
            "devmode",
            "local_context_focus",
            "SRD",
            "pretrained_model_name",
            "state_dict",
            "single_targets",
            "multi_targets",
            "targetclasses",
            "knowledgesources",
            "is_use_natural_target_phrase_for_spc",
            "default_lm",
            "coref_mode_in_training",
        ]

        combinations = None
        if self.opt.combi_mode == "default":
            combinations = combinations_default_0
        elif self.opt.combi_mode == "top":
            combinations = combinations_top
        elif self.opt.combi_mode == "top_coref8":
            combinations = combinations_top_coref8
        elif self.opt.combi_mode == "top_coref8_stanceno":
            combinations = combinations_top_coref8_stanceno
        elif self.opt.combi_mode == "top_coref16":
            combinations = combinations_top_coref16
        elif self.opt.combi_mode == "top_finetuned":
            combinations = combinations_top_finetuned
        elif self.opt.combi_mode == "top4otherdomains":
            combinations = combinations_top4otherdomains
        elif self.opt.combi_mode == "ensemble":
            combinations = combinations_ensemble
        elif self.opt.combi_mode == "final_stance0":
            combinations = combinations_final_stance0
        elif self.opt.combi_mode == "final_stance1":
            combinations = combinations_final_stance1
        elif self.opt.combi_mode == "final_nostance":
            combinations = combinations_final_nostance
        elif self.opt.combi_mode == "final_tdbert":
            combinations = combinations_final_tdbert
        elif self.opt.combi_mode == "final_lcf2":
            combinations = combinations_final_lcf2
        elif self.opt.combi_mode == "final_spc":
            combinations = combinations_final_spc

        if not combinations:
            raise ValueError(
                "combination(mode={}, id={}) not defined".format(
                    self.opt.combi_mode, self.opt.combi_id
                )
            )

        # key: name of parameter that is only applied if its conditions are met
        # pad_value: list of tuples, consisting of parameter name and the pad_value it needs to have in order for the
        # condition to be satisfied
        # Note that all tuples in this list are OR connected, so if at least one is satisfied, the conditions are met.
        # If we need AND connected conditions, my idea is to add an outer list, resulting in a list of lists (of
        # tuples) where all lists are AND connected.
        # If a condition is not satisfied, the corresponding parameter will still be pass
        # if a key is not listed here, there are no conditions and thus the key / parameter will be used always
        conditions = {
            "is_use_natural_target_phrase_for_spc": [("own_model_name", "spc_bert")],
            "spc_lm_representation": [("own_model_name", "spc_bert"),],
            "spc_input_order": [("own_model_name", "spc_bert"),],
            "aen_lm_representation": [("own_model_name", "aen_bert"),],
            "use_early_stopping": [("num_epoch", "10"),],
            "local_context_focus": [
                ("own_model_name", "lcf_bert"),
                ("own_model_name", "lcf_bert2"),
                ("own_model_name", "lcfs_bert"),
                ("own_model_name", "lcfst_bert"),
            ],
            "SRD": [
                ("own_model_name", "lcf_bert"),
                ("own_model_name", "lcf_bert2"),
                ("own_model_name", "lcfs_bert"),
                ("own_model_name", "lcfst_bert"),
            ],
            "single_targets": [
                # we need single_targets mode only for models that are single target
                ("own_model_name", "lcf_bert"),
                ("own_model_name", "lcf_bert2"),
                ("own_model_name", "lcfs_bert"),
                ("own_model_name", "lcft_bert"),
                ("own_model_name", "lcfst_bert"),
                ("own_model_name", "aen_bert"),
                ("own_model_name", "spc_bert"),
                ("own_model_name", "tdbert"),
                ("own_model_name", "EnsembleTopA"),
                ("own_model_name", "EnsembleTopB"),
                ("own_model_name", "tdbert-qa-mul"),
                ("own_model_name", "tdbert-qa-con"),
                ("own_model_name", "tdbertlikesingle"),
                ("own_model_name", "notargetclsbert"),
                ("own_model_name", "random_single"),
                ("own_model_name", "grutsc"),
            ],
            "multi_targets": [
                ("own_model_name", "tdbertlikemulti"),
                ("own_model_name", "tdbertlikemulti_dense"),
                ("own_model_name", "seq2seq"),
                ("own_model_name", "seq2seq_withouttargetmask"),
                ("own_model_name", "random_multi"),
                ("own_model_name", "contrasting"),
            ],
            "knowledgesources": [
                ("own_model_name", "EnsembleTopA"),
                ("own_model_name", "EnsembleTopB"),
                ("own_model_name", "grutsc"),
            ],
            "pretrained_model_name": [("default_lm", "bert-base-uncased"),],
        }

        assert set(args_names_ordered) == set(combinations.keys()), (
            f"mismatch args_names_ordered vs. combinations: "
            f"{set(args_names_ordered).symmetric_difference(combinations)}"
        )

        self.experiment_base_id = (
            self.opt.dataset + "_" + datetime.today().strftime("%Y%m%d-%H%M%S")
        )
        self.basecmd = ["python", "train.py"]
        self.basepath = "controller_data"
        self.basepath_data = os.path.join(self.basepath, "datasets")

        combination_count = 1
        _combination_values = []
        for arg_name in args_names_ordered:
            arg_values = list(combinations[arg_name])
            combination_count = combination_count * len(arg_values)
            _combination_values.append(arg_values)

        combinations = list(product(*_combination_values))
        assert len(combinations) == combination_count

        self.logger.info(
            "{} arguments, totaling in {} combinations".format(
                len(args_names_ordered), combination_count
            )
        )

        # apply conditions
        self.logger.info("applying conditions...")
        self.named_combinations, count_duplicates = self._apply_conditions(
            combinations, args_names_ordered, conditions
        )
        self.logger.info(
            "applied conditions. removed {} combinations. {} -> {}".format(
                count_duplicates, combination_count, len(self.named_combinations)
            )
        )
        self.combination_count = len(self.named_combinations)

        # expand by number of runs per experiment
        expanded_named_combinations = []
        for named_combination in self.named_combinations:
            for run_id in range(self.opt.num_runs_per_experiment):
                expanded_named_combination = named_combination.copy()
                expanded_named_combination["run_id"] = run_id
                expanded_named_combinations.append(expanded_named_combination)
        self.logger.info(
            "expanded to %s runs per experiment. combination count: %s -> %s",
            self.opt.num_runs_per_experiment,
            self.combination_count,
            len(expanded_named_combinations),
        )
        self.named_combinations = expanded_named_combinations
        self.combination_count = len(self.named_combinations)

        if self.opt.dataset == "semeval14restaurants":
            (
                self.dataset_preparer,
                self.datasetname,
                self.task_format,
            ) = DatasetPreparer.semeval14restaurants(self.basepath_data)
        elif self.opt.dataset == "semeval14laptops":
            (
                self.dataset_preparer,
                self.datasetname,
                self.task_format,
            ) = DatasetPreparer.semeval14laptops(self.basepath_data)
        elif self.opt.dataset == "acl14twitter":
            (
                self.dataset_preparer,
                self.datasetname,
                self.task_format,
            ) = DatasetPreparer.acl14twitter(self.basepath_data)
        elif (
            self.opt.dataset
            == "newsmtsc_devtest_rw"
        ):
            (
                self.dataset_preparer,
                self.datasetname,
                self.task_format,
            ) = DatasetPreparer.newsmtsc_devtest_rw(
                self.basepath_data
            )
        elif (
                self.opt.dataset
                == "newsmtsc_devtest_mt"
        ):
            (
                self.dataset_preparer,
                self.datasetname,
                self.task_format,
            ) = DatasetPreparer.newsmtsc_devtest_mt(
                self.basepath_data
            )
        else:
            raise Exception("unknown dataset: {}".format(self.opt.dataset))

    def _apply_conditions(self, combinations, args_names_ordered, conditions):
        named_combinations = []
        seen_experiment_ids = set()
        count_duplicates = 0

        with tqdm(total=len(combinations)) as pbar:
            for combination in combinations:
                named_combination = {}
                full_named_combination = self._args_combination_to_single_arg_values(
                    combination, args_names_ordered
                )

                # for a parameter combination, pass only those parameters that are valid
                # for that combination
                for arg_index, arg_name in enumerate(args_names_ordered):
                    # iterate each parameter and validate - using the other parameter
                    # names and values - whether its conditions are met
                    if self._check_conditions(
                        arg_name, full_named_combination, conditions
                    ):
                        # if yes, pass it
                        named_combination[arg_name] = combination[arg_index]
                        self.logger.debug(
                            "using '{}' in combination {}".format(arg_name, combination)
                        )
                    else:
                        self.logger.debug(
                            "not using '{}' in combination {}".format(
                                arg_name, combination
                            )
                        )

                # check if experiment_id of named combination was already seen
                experiment_id = self._experiment_named_id_from_named_combination(
                    named_combination
                )
                if experiment_id not in seen_experiment_ids:
                    seen_experiment_ids.add(experiment_id)
                    named_combinations.append(named_combination)
                else:
                    count_duplicates += 1

                pbar.update(1)

        return named_combinations, count_duplicates

    def _check_conditions(self, arg_name, full_named_combination, conditions):
        """
        For a given parameter, checks whether its conditions are satisfied. If so, returns True, else False.
        :param arg_name:
        :param arg_value:
        :return:
        """
        if arg_name in conditions and len(conditions[arg_name]) >= 1:
            # at this point we know that there are conditions for the given parameters
            or_connected_conditions = conditions[arg_name]

            for cond_tup in or_connected_conditions:
                cond_param_name = cond_tup[0]
                cond_param_value = cond_tup[1]

                # get parameter and its pad_value in current combination
                if full_named_combination[cond_param_name] == cond_param_value:
                    return True

            # since there was at least one condition due to our check above, we return
            # False here, since the for loop did not return True
            return False

        else:
            # if there is no condition associated with arg_name just return true
            return True

    def _build_args(self, named_args):
        args_list = []
        for arg_name, arg_val in named_args.items():
            self._add_arg(args_list, arg_name, arg_val)

        return args_list

    def _add_arg(self, args_list, name, value):
        args_list.append("--" + name)
        args_list.append(str(value))
        return args_list

    def _prepare_experiment_env(self, experiment_path):
        os.makedirs(experiment_path, exist_ok=True)
        self.dataset_preparer.export(os.path.join(experiment_path, "datasets"))

    def _args_combination_to_single_arg_values(
        self, args_combination, args_names_ordered
    ):
        args_names_values = {}
        for arg_index, arg_name in enumerate(args_names_ordered):
            args_names_values[arg_name] = args_combination[arg_index]
        return args_names_values

    def _experiment_named_id_from_named_combination(self, named_combination):
        return "__".join(["{}={}".format(k, v) for (k, v) in named_combination.items()])

    def prepare_single_setup(self, named_combination, experiment_id, pool_index):
        experiment_path = "{}/{}/".format(self.experiment_base_id, experiment_id)
        experiment_path = os.path.join(self.experiment_base_path, experiment_path)

        self._prepare_experiment_env(experiment_path)

        args = self._build_args(named_combination)
        self._add_arg(args, "snem", self.snem)
        self._add_arg(args, "dataset_name", self.datasetname)
        self._add_arg(args, "experiment_path", experiment_path)
        self._add_arg(args, "data_format", self.task_format)

        if self.cuda_devices:
            # we use the pool_index to get the respective pool's cuda device
            cuda_device_id = self.cuda_devices[pool_index]
            cuda_device_name = "cuda:" + str(cuda_device_id)
            self._add_arg(args, "device", cuda_device_name)

        cmd = self.basecmd + args
        human_cmd = " ".join(cmd)

        with open(os.path.join(experiment_path, "experiment_cmd.sh"), "w") as writer:
            writer.write(human_cmd)

        return cmd, human_cmd, experiment_path

    def run(self):
        global completed_tasks

        if not self.opt.results_path:
            results_path = "results/default_results_{}".format(self.datasetname)
        else:
            results_path = self.opt.results_path

        if not self.opt.continue_run:
            self.logger.info("not continuing")
            os.remove(results_path)
        else:
            self.logger.info("continuing previous run(s)")

        completed_tasks = DiskDict(results_path)
        self.logger.info("found {} previous results".format(len(completed_tasks)))

        self.logger.info("preparing experiment setups...")
        experiment_descs = defaultdict(list)
        previous_tasks = Counter()
        with tqdm(total=self.combination_count) as pbar:
            for i, named_combination in enumerate(self.named_combinations):
                _experiment_named_id = self._experiment_named_id_from_named_combination(
                    named_combination
                )
                # i modulo the number of workers determines which pool this experiment
                # will run in. the pool_index will also later be used to determine
                # which cuda device (if any) to use, thereby making sure, that all tasks
                # of one pool use the same device, whereas each device will be used
                # solely by only one pool (this way, we make sure that at no time two
                # or more tasks will run that use the same cuda device)
                pool_index = i % self.opt.num_workers

                if _experiment_named_id in completed_tasks:
                    task_desc = completed_tasks[_experiment_named_id]

                    if task_desc["rc"] != 0:
                        if self.opt.rerun_non_rc0:
                            self.logger.debug(
                                "task {} was already executed, "
                                "but with rc={}. rerunning.".format(
                                    _experiment_named_id, task_desc["rc"]
                                )
                            )
                            (
                                cmd,
                                human_cmd,
                                experiment_path,
                            ) = self.prepare_single_setup(
                                named_combination, i, pool_index
                            )

                            experiment_descs[pool_index].append(
                                (
                                    i,
                                    _experiment_named_id,
                                    named_combination,
                                    cmd,
                                    human_cmd,
                                    experiment_path,
                                )
                            )

                            previous_tasks["new"] += 1
                            del completed_tasks[_experiment_named_id]
                        else:
                            self.logger.debug(
                                "task {} was already executed, but with rc={}. not "
                                "rerunning.".format(
                                    _experiment_named_id, task_desc["rc"]
                                )
                            )
                            previous_tasks["rcnon0"] += 1
                    else:
                        # rerun tasks where the rc != 0 (always rerun tasks that have
                        # not been executed at all, yet)
                        self.logger.debug(
                            "skipping experiment: {}".format(_experiment_named_id)
                        )
                        self.logger.debug(
                            "previous result: {}".format(
                                completed_tasks[_experiment_named_id]
                            )
                        )
                        previous_tasks["rc0"] += 1
                else:
                    (cmd, human_cmd, experiment_path,) = self.prepare_single_setup(
                        named_combination, i, pool_index
                    )

                    experiment_descs[pool_index].append(
                        (
                            i,
                            _experiment_named_id,
                            named_combination,
                            cmd,
                            human_cmd,
                            experiment_path,
                        )
                    )
                    previous_tasks["new"] += 1
                pbar.update(1)

        self.logger.info(
            "summary (new is also increased for tasks that were executed previously "
            "but yielded rc!=0 - if rerun_non_rc0==True"
        )
        self.logger.info("{}".format(previous_tasks))

        self.logger.info("starting {} experiments".format(self.combination_count))
        self.logger.info(
            "creating {} process pools with each 1 worker".format(self.opt.num_workers)
        )

        if self.cuda_devices and len(self.cuda_devices) != self.opt.num_workers:
            self.logger.warning(
                "number of cuda devices does not match number of workers: "
                "{} vs. {}".format(len(self.cuda_devices), self.opt.num_workers)
            )

        manager = multiprocessing.Manager()
        running_processes = manager.dict()
        for pool_index in range(self.opt.num_workers):
            pool = multiprocessing.Pool(processes=1, maxtasksperchild=1)
            for desc in experiment_descs[pool_index]:
                proc_args = desc + (running_processes,)  # must be a tuple
                pool.apply_async(
                    start_worker,
                    args=proc_args,
                    callback=on_task_done,
                    error_callback=on_task_error,
                )
            self.logger.info(
                f"created pool with {len(experiment_descs[pool_index])} tasks, each "
                f"processed by 1 worker"
            )

        self.logger.info("waiting for workers to complete all jobs...")
        prev_count_done = 0
        with tqdm(total=previous_tasks["new"], initial=prev_count_done) as pbar:
            while completed_tasks_in_this_run_count < previous_tasks["new"]:
                time.sleep(10)
                update_inc = completed_tasks_in_this_run_count - prev_count_done

                if update_inc > 0:
                    pbar.update(update_inc)
                    prev_count_done = completed_tasks_in_this_run_count
                    completed_tasks.sync_to_disk()

                best_dev_snem, best_experiment_id = self._get_best_dev_snem()
                sorted_running_procs = sorted(
                    self._get_running_processes(running_processes),
                    key=lambda running_device: running_device[1],
                )
                pbar.set_postfix_str(
                    f"dev-snem: {best_dev_snem:.3f} ({best_experiment_id}); prcs: {sorted_running_procs}"
                )

        # save to disk
        completed_tasks.sync_to_disk()
        self.logger.info(
            f"finished all tasks ({completed_tasks_in_this_run_count} of "
            f"{previous_tasks['new']})"
        )

        # clone the results so that we can process them while not modifying the original
        processed_results = dict(completed_tasks)

        experiments_rc_overview = Counter()
        non_okay_experiment_ids = []
        for experiment_named_id, experiment_result in processed_results.items():
            rc = experiment_result["rc"]
            experiments_rc_overview[rc] += 1

            if rc != 0:
                if rc == 99:
                    pass
                else:
                    non_okay_experiment_ids.append(experiment_result["experiment_id"])

        self.logger.warning(
            f"{len(non_okay_experiment_ids)} experiments did not return 0: "
            f"{sorted(non_okay_experiment_ids)}"
        )

        # snem-based performance sort
        sorted_results = list(dict(processed_results).values())
        for result in sorted_results:
            if result["details"]:
                result["dev_snem"] = result["details"]["dev_stats"][self.snem]
                del result["details"]
            else:
                result["dev_snem"] = -1.0
        sorted_results.sort(key=lambda x: x["dev_snem"], reverse=True)
        headers = list(sorted_results[0].keys())
        rows = [x.values() for x in sorted_results]

        self.logger.info("all experiments finished. statistics:")
        self.logger.debug("snem-based performances:")
        self.logger.debug("\n" + tabulate(rows, headers))
        self.logger.info("return codes: {}".format(experiments_rc_overview))
        self.logger.info(
            "best dev-snem: %s (id: %s)",
            sorted_results[0]["dev_snem"],
            sorted_results[0]["experiment_id"],
        )

    def _get_best_dev_snem(self):
        best_dev_snem = -1.0
        experiment_id = -1
        # to avoid that the dict is modified while iterating it, copy it for temporary use
        # if we're unlucky this could happen otherwise: RuntimeError: dictionary changed size during iteration
        tmp_tasks = completed_tasks.copy()
        for task in tmp_tasks.values():
            if task.get("details") and task["details"].get("dev_stats"):
                best_dev_snem = max(
                    best_dev_snem, task["details"]["dev_stats"][self.snem]
                )
                experiment_id = task["experiment_id"]
        return best_dev_snem, experiment_id

    def _get_running_processes(self, running_processes):
        return [
            (exp_id, running_device[1])
            for exp_id, running_device in running_processes.items()
            if running_device[0]
        ]


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--dataset", default=None, type=str)
    parser.add_argument("--experiments_path", default="./experiments", type=str)
    parser.add_argument(
        "--continue_run", type=str2bool, nargs="?", const=True, default=True
    )
    parser.add_argument(
        "--rerun_non_rc0", type=str2bool, nargs="?", const=True, default=True
    )
    parser.add_argument("--num_workers", type=int, default=1)
    parser.add_argument("--results_path", type=str, default=None)
    parser.add_argument("--combi_mode", type=str, required=True)
    parser.add_argument("--combi_id", type=int, default=0)
    parser.add_argument(
        "--cuda_devices",
        type=str,
        default=None,
        help="Comma separated list of cuda device IDs, e.g., "
        "0,1,2,3; or SGE_GPU to read this list from "
        "the environment variable SGE_GPU. If not defined,"
        "CPU will be used instead of GPU.",
    )
    parser.add_argument("--num_runs_per_experiment", type=int, default=1)
    parser.add_argument(
        "--only_calc_combinations", type=str2bool, nargs="?", const=True, default=False
    )
    opt = parser.parse_args()

    if opt.only_calc_combinations:
        SetupController(opt)
    else:
        SetupController(opt).run()



from argparse import Namespace
from typing import Dict

import torch.nn as nn

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.models.FXBaseModel import FXBaseModel


class TDBertLikeMultiTargetDense(FXBaseModel):
    """
    This model returns a sequence that only contains the hidden states of those output
    nodes that represent a word piece of the target phrase. All other hidden states are
    set to 0. The length of the output sequence is opt.max_seq_len

    From a conceptual perspective, TD-BERT is in some aspects similar (e.g., only the
    output of target-phrase-related nodes is used), but in other aspects not similar or
    it is not quite clear how the authors of TD-BERT implemented them. An email I sent
    to them was not answered yet.
    """

    @staticmethod
    def get_language_models():
        return (BERT_BASE_UNCASED,)

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(TDBertLikeMultiTargetDense, self).__init__()
        self.language_model = transformer_models[BERT_BASE_UNCASED]
        self.dropout = nn.Dropout(opt.dropout)
        self.dense = nn.Bilinear(
            self.language_model.config.hidden_size,
            FXDataset.NUM_MAX_TARGETS_PER_ITEM,
            FXDataset.NUM_MAX_TARGETS_PER_ITEM,
        )
        self.dense2 = nn.Linear(opt.max_seq_len, opt.polarities_dim)

    def forward(self, inputs):
        # get inputs
        text_bert_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        text_bert_indices_targets_mask = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK
        )
        # prepare inputs
        # for text only, we do not need target specific information, i.e., all text
        # vectors are identical. also, bert, can
        # only process one sequence of size max_seq_len (more specifically, a tensor
        # of size batch_size x max_seq_ken). thus, we select only the first element from
        # the second dimension (the dimensions are: batch, targets, hidden_states)
        text_bert_indices = text_bert_indices[:, 0, :]
        # apply bert
        last_hidden_states, pooler_output, all_hidden_states = self.language_model(
            input_ids=text_bert_indices
        )

        # dropout
        last_hidden_states = self.dropout(last_hidden_states)

        # shapes:
        # last_hidden_states:                   batch, seqlen, bertdim
        # text_bert_indices_targets_mask:       batch, target, seqlen
        # new text_bert_indices_targets_mask:   batch, seqlen, target
        text_bert_indices_targets_mask = text_bert_indices_targets_mask.permute(
            0, 2, 1
        ).clone()

        logits = self.dense(last_hidden_states, text_bert_indices_targets_mask)
        logits = logits.permute(0, 2, 1)
        logits = self.dense2(logits)

        return logits



import random
from collections import Counter
from typing import List, Iterable, Set

import jsonlines
import networkx as nx
import numpy as np
import spacy
import torch
from imblearn.over_sampling import RandomOverSampler
from torch.utils.data import Dataset
from tqdm import tqdm
from transformers import (
    BertTokenizer,
    RobertaTokenizer,
    XLNetTokenizer,
    AlbertTokenizer,
)

from NewsSentiment.SentimentClasses import SentimentClasses
from NewsSentiment.customexceptions import TooLongTextException, TargetNotFoundException
from NewsSentiment.consts import *
from NewsSentiment.fxlogger import get_logger

# get logger
from NewsSentiment.knowledge.bingliuopinion.bingliuopinion import (
    get_bingliu_polarities_as_tensor,
    get_num_bingliu_polarities,
)
from NewsSentiment.knowledge.mpqasubjectivity.mpqasubjectivity import (
    get_mpqa_subjectivity_polarities_as_tensor,
    get_num_mpqa_subjectivity_polarities,
)
from NewsSentiment.knowledge.nrcemolex.nrcemolex import (
    get_nrc_emotions_as_tensor,
    get_num_nrc_emotions,
)
from NewsSentiment.knowledge.zeros.zerosknowledge import get_num_zero_dimensions, get_zeros_as_tensor
from NewsSentiment.models.FXBaseModel import FXBaseModel

logger = get_logger()
try:
    nlp = spacy.load("en_core_web_sm")
except OSError:
    spacy.cli.download("en_core_web_sm")
    nlp = spacy.load("en_core_web_sm")
# this is what we could run on spacy 2.2+
# nlp_dep_parser_labels = list(nlp.parser.labels)
# but since we're on spacy 2.1 (and need to be because of newsalyze-backend)
# we have to use this workaround, see:
# https://github.com/explosion/spaCy/discussions/5135
nlp_dep_parser_labels = []
for move in nlp.parser.move_names:
    if "-" in move:
        move = move.split("-")[1]
        if "||" in move:
            move = move.split("||")[1]
        nlp_dep_parser_labels.append(move)
# assert set(nlp_dep_parser_labels) == {
#     "agent",
#     "acl",
#     "intj",
#     "predet",
#     "expl",
#     "prt",
#     "meta",
#     "attr",
#     "pcomp",
#     "parataxis",
#     "compound",
#     "auxpass",
#     "mark",
#     "prep",
#     "dative",
#     "xcomp",
#     "npadvmod",
#     "advcl",
#     "preconj",
#     "pobj",
#     "csubj",
#     "dep",
#     "punct",
#     "case",
#     "ROOT",
#     "quantmod",
#     "cc",
#     "det",
#     "subtok",
#     "neg",
#     "nmod",
#     "amod",
#     "advmod",
#     "nsubjpass",
#     "appos",
#     "conj",
#     "dobj",
#     "acomp",
#     "nsubj",
#     "csubjpass",
#     "nummod",
#     "ccomp",
#     "oprd",
#     "poss",
#     "relcl",
#     "aux",
# }


class RandomOversampler(torch.utils.data.sampler.Sampler):
    def __init__(self, dataset: Dataset, random_seed=None):
        x = []
        y = []
        for ind, example in enumerate(dataset):
            x.append(ind)
            y.append(example["polarity"])

        x_arr = np.asarray(x).reshape((len(x), 1))
        y_arr = np.asarray(y).ravel()

        ros = RandomOverSampler(random_state=random_seed)
        x_sampled, y_sampled = ros.fit_resample(x_arr, y_arr)
        self.sampled_indexes = x_sampled.ravel().tolist()
        sampled_labels = y_sampled.tolist()

        assert len(self.sampled_indexes) == len(sampled_labels)

        random.shuffle(self.sampled_indexes)

        get_logger().info(
            f"oversampled to {len(self.sampled_indexes)} samples. label distribution: "
            f"{Counter(sampled_labels)}"
        )

    def __len__(self):
        return len(self.sampled_indexes)

    def __iter__(self):
        return iter(self.sampled_indexes)


class FXEasyTokenizer:
    NUM_CATEGORIES_OF_SELECTED_KNOWLEDGE_SOURCES = 0
    __PROCESSED_KNOWLEDGE_SOURCES = set()

    def __init__(
        self,
        tokenizers_name_and_obj: dict,
        max_seq_len: int,
        knowledge_sources: Iterable[str],
        is_use_natural_target_phrase_for_spc: bool,
    ):
        self.tokenizers_name_and_obj = tokenizers_name_and_obj
        self.max_seq_len = max_seq_len
        self.knowledge_sources = knowledge_sources
        self.is_use_natural_target_phrase_for_spc = is_use_natural_target_phrase_for_spc

    @staticmethod
    def create_entire_text(
        text_left: str,
        target_phrase: str,
        text_right: str,
        is_return_modified_text_left: bool,
    ):
        """
        Create the entire text from the three prepared text segment. Will modify
        text_left (append a whitespace " ") if it is empty, which is necessary for
        target mask creation.
        :param is_return_modified_text_left:
        :param text_left:
        :param target_phrase:
        :param text_right:
        :return:
        """
        # input text must have basic tokenization, i.e., each token is separated by ' '
        if len(text_left) > 0:
            assert text_left[-1] == " ", f"left is not prepared: '{text_left[-1]}'"
        else:
            # we need to have a non-empty left sequence, otherwise BERT's tokenizer
            # will raise an exception
            text_left = " "
        if len(text_right) > 0:
            assert text_right[0] == " ", f"right is not prepared: '{text_right[0]}'"

        # produce entire text string
        text = text_left + target_phrase + text_right

        if is_return_modified_text_left:
            return text, text_left
        else:
            return text

    @staticmethod
    def prepare_left_segment(text_left: str):
        """
        Prepares the left text segment. See FXDataset:task_to_dataset_item for more
        information. Removes any leading newlines.
        :param text_left:
        :return:
        """
        while len(text_left) > 0 and text_left[0] == "\n":
            text_left = text_left[1:]
        if len(text_left) > 0 and text_left[-1] != " ":
            text_left += " "
        return text_left

    @staticmethod
    def prepare_target_mention(target_mention: str):
        """
        Removes any leading newlines.
        :param target_mention:
        :return:
        """
        while len(target_mention) > 0 and target_mention[0] == "\n":
            target_mention = target_mention[1:]
        return target_mention

    @staticmethod
    def prepare_right_segment(text_right: str):
        """
        Prepares the right text segment. See FXDataset:task_to_dataset_item for more
        information.
        :param text_right:
        :return:
        """
        if len(text_right) > 0 and text_right[0] != " ":
            text_right = " " + text_right
        return text_right

    def _create_word_to_wordpiece_mapping(
        self, tokenizer, words: List, for_text_with_special_tokens, offset
    ):
        if not for_text_with_special_tokens:
            raise NotImplementedError()

        previous_words = []
        mapping = None
        # offset word_index by 1 (0th word has index 1)
        # words_without_single_whitespace = [word for word in words if word != " "]
        for word_index, word in enumerate(words, start=offset):
            # update left and previous words
            left = " ".join(previous_words) + " "
            previous_words.append(word)

            # produce target masks
            target_mask = self._create_target_mask(
                tokenizer=tokenizer,
                text_left=left,
                target=word,
                text_right="",
                for_text_with_special_tokens=for_text_with_special_tokens,
                is_raise_exception_if_target_after_max_seq_len=False,
            )

            # if the current word is after the max seq len, the target mask = "toolong"
            if target_mask == "toolong":
                # we abort the for loop since all remaining words will also be
                # after the max seq len
                break

            target_mask = torch.LongTensor(target_mask)
            target_mask = torch.mul(target_mask, word_index)
            if mapping is None:
                mapping = target_mask
            else:
                # make sure there is no overlap
                if type(tokenizer) == RobertaTokenizer:
                    if sum(mapping * target_mask) > 0:
                        logger.debug(
                            "overlap when mapping tokens to wordpiece (allow overwriting because "
                            "Roberta is used)"
                        )
                else:
                    assert sum(mapping * target_mask) == 0
                mapping = mapping + target_mask

        assert mapping.shape[0] <= self.max_seq_len
        return mapping, previous_words

    def _calculate_dep_matrix(self, text_tokens, text_tokens_as_str):
        _check_len_doc = len(text_tokens)
        assert _check_len_doc == len(
            text_tokens_as_str
        ), f"{_check_len_doc} vs {len(text_tokens_as_str)}: {text_tokens} vs {text_tokens_as_str}"

        # for the dependency matrix
        dependency_tensor = torch.zeros(
            len(text_tokens_as_str), len(text_tokens_as_str), dtype=torch.long
        )

        for token_index, token in enumerate(text_tokens):
            if token.is_space:
                # text_tokens_as_str does not contain whitespace tokens, so we skip them
                # here, too
                continue
            assert (
                token.text == text_tokens_as_str[token_index]
            ), f"'{token.text}' vs. '{text_tokens_as_str[token_index]}'"

            # to build the dependency relation matrix
            # get the token's head and type of relation
            head_of_token = token.head
            try:
                index_in_token_list_of_head_of_token = text_tokens.index(head_of_token)
            except ValueError as ve:
                # if the head cannot be found in the (truncated to wordpiece equivalent
                # max seq len) list of text_tokens, this means that the current token's
                # head is not part of the wordpiece sequence. we print an error and
                # not change the current word's vector (will be all 0's)
                logger.warning(ve)
                continue

            relation_to_head = token.dep_
            # offset the relation by 1 so that the root relation (which is 0) is
            # non-zero
            index_of_relation_to_head = (
                nlp_dep_parser_labels.index(relation_to_head) + 1
            )
            # insert to dependency tensor
            dependency_tensor[
                token_index, index_in_token_list_of_head_of_token
            ] = index_of_relation_to_head

        return dependency_tensor

    def _calculate_dep_distance(self, text_tokens, text_left_len, target):
        threshold_for_target_token = 8

        # by using spacys tokenization also on the target (instead of a simple
        # whitespace split as in https://github.com/StevePhan101/LCFS-BERT/
        # we ensure that the same tokenization as was used for the text is applied for
        # the target
        nlp_target = nlp(target)
        # target_terms_lowercased = [a.lower() for a in target.split()]
        target_terms_lowercased = [a.text.lower() for a in nlp_target]

        # Load spacy's dependency tree into a networkx graph
        edges = []
        cnt = 0
        found_start_of_target = False
        target_term_ids = [0] * len(target_terms_lowercased)
        len_previously_processed_tokens = 0
        for token in text_tokens:
            # Record the position of aspect terms
            lowercased_token = token.lower_
            token_startchar_indoc = token.idx
            token_endchar_indoc = token_startchar_indoc + len(token.text_with_ws)

            if (
                cnt < len(target_terms_lowercased)
                and lowercased_token == target_terms_lowercased[cnt]
            ):
                # if both conditions above are satisfied, we probably have (part of)
                # the target in the current token. however, to be slightly more
                # confident about this, also ensure that the previous tokens have
                # approximately the same length as text_left_len. or skip that check
                # if we already found the start of the target
                if (
                    found_start_of_target
                    or abs(text_left_len - len_previously_processed_tokens)
                    < threshold_for_target_token
                ):
                    found_start_of_target = True
                    target_term_ids[cnt] = token.i
                    cnt += 1
                else:
                    pass

            len_previously_processed_tokens += len(token.text_with_ws)

            # for lcfs absa
            for child in token.children:
                edges.append(
                    (
                        "{}_{}".format(token.lower_, token.i),
                        "{}_{}".format(child.lower_, child.i),
                    )
                )

        if not found_start_of_target:
            raise TargetNotFoundException(
                f"no target found: {text_tokens}, {text_left_len}, {target}"
            )

        # create the graph
        graph = nx.Graph(edges)

        dist = [0.0] * len(text_tokens)
        for i, word in enumerate(text_tokens):
            source = "{}_{}".format(word.lower_, word.i)
            sum = 0
            for term_id, term in zip(target_term_ids, target_terms_lowercased):
                target = "{}_{}".format(term, term_id)
                try:
                    sum += nx.shortest_path_length(graph, source=source, target=target)
                except:
                    sum += len(text_tokens)  # No connection between source and target
            dist[i] = sum / len(target_terms_lowercased)
        return dist

    def _create_target_mask(
        self,
        tokenizer,
        text_left: str,
        target: str,
        text_right: str,
        for_text_with_special_tokens: bool,
        is_raise_exception_if_target_after_max_seq_len: bool = True,
    ):
        """
        Creates a target mask for a single target in a text that comprises left, target,
        and right. Specifically, the target mask has the same length as the sequence of
        the entire text including special tokens, and contains 1's for each position k
        where in the entire text sequence (including special tokens) the token at
        position k is part of the target phrase.
        Simplified example (this does not consider any LM specific tokenization
        characteristics, e.g., most LMs' tokenizers do not tokenize single tokens but
        produce sub-tokens (e.g., BERT using the WordPiece tokenizer).
        Text:                                    Mr. <target>Smith</target> is great .
        Text indexes (with special tokens):  CLS 12           50            18 15    19 END
        Target mask:                          0  0             1             0  0    0   0
        :param tokenizer:
        :param text_left:
        :param target:
        :param text_right:
        :param text_pair: if not None, will be appended to the input sequence as the second pair
        :return:
        """
        if not for_text_with_special_tokens:
            raise NotImplementedError()
        assert target != " ", "passed a single whitespace as target"

        text, text_left = self.create_entire_text(
            text_left, target, text_right, is_return_modified_text_left=True
        )

        # get token ids, cf. https://huggingface.co/transformers/glossary.html
        text_left_ids_with_special_tokens = tokenizer.encode(
            text_left, add_special_tokens=True
        )
        text_right_ids = tokenizer.encode(text_right, add_special_tokens=False)
        text_ids_with_special_tokens = tokenizer.encode(text, add_special_tokens=True)
        target_phrase_ids_with_special_tokens = tokenizer.encode(
            target, add_special_tokens=True
        )

        len_text_left_ids_with_special_tokens = len(text_left_ids_with_special_tokens)
        len_text_right_ids = len(text_right_ids)
        len_text_id_with_special_tokens = len(text_ids_with_special_tokens)
        len_target_ids_with_special_tokens = len(target_phrase_ids_with_special_tokens)

        token_obj_class = type(tokenizer)
        if token_obj_class in [BertTokenizer, RobertaTokenizer, AlbertTokenizer]:
            # these tokenizers create 1 special token at the end that should not be
            num_special_tokens_at_start = 1
            num_special_tokens_at_end = 1
        elif token_obj_class == XLNetTokenizer:
            num_special_tokens_at_start = 0
            num_special_tokens_at_end = 2
        else:
            raise ValueError
        # determine numbers of special tokens in the target
        num_special_tokens_in_target = (
            num_special_tokens_at_start + num_special_tokens_at_end
        )
        # get the number of token ids of the target, excluding special tokens
        # we do this, because
        num_effective_length_target_in_text = (
            len_target_ids_with_special_tokens - num_special_tokens_in_target
        )
        # get the position of the first non-special token of the target phrase in the
        # sequence of tokens representing the entire text, which includes special tokens
        tp_seq_in_text_seq_pos_start = (
            len_text_left_ids_with_special_tokens - num_special_tokens_at_end
        )
        # likewise, get the position of the last non-special token of the target phrase
        # in the sequences of tokens representing the entire text, which includes
        # special tokens
        tp_seq_in_text_seq_pos_end = (
            tp_seq_in_text_seq_pos_start + num_effective_length_target_in_text
        )
        # get the length of the token sequence representing the target phrase (again,
        # only actual content tokens, no special tokens)
        tp_seq_in_text_seq_len = (
            tp_seq_in_text_seq_pos_end - tp_seq_in_text_seq_pos_start
        )
        assert (
            tp_seq_in_text_seq_len > 0
        ), f"tp_seq_in_text_seq_len: {tp_seq_in_text_seq_len}"

        # if the end position of the target is greater than the maximum sequence length,
        # we raise an Exception, since it won't make sense to train with or evaluate on
        # this example since the target will be out of bounds
        if tp_seq_in_text_seq_pos_end > self.max_seq_len:
            if is_raise_exception_if_target_after_max_seq_len:
                raise TooLongTextException()
            else:
                # e.g., if invoked for creating the target mapping
                return "toolong"

        # create target mask: all 0 except for occurrence of target
        target_mask_seq = [0] * self.max_seq_len
        target_mask_seq[tp_seq_in_text_seq_pos_start:tp_seq_in_text_seq_pos_end] = [
            1
        ] * tp_seq_in_text_seq_len

        # perform checks
        count_nonzero_elements_in_target_mask_seq = sum(
            [1 if elem != 0 else 0 for elem in target_mask_seq]
        )

        # roberta: 1 special token at beginning, 1 at end
        # bert: 1 beginning, 1 at end
        # xlnet: 0 beginning, 2 at end
        # len_text_left_ids_with_special_tokens contains the right amount of both
        # special tokens at the beginning and end of the sequence
        calculated_len = (
            len_text_left_ids_with_special_tokens
            + num_effective_length_target_in_text
            + len_text_right_ids
        )

        if token_obj_class == RobertaTokenizer:
            # Roberta has more complex tokenization behavior, e.g., >>> tokenizer("the friendly staff welcomes")['input_ids']
            # [0, 627, 5192, 813, 16795, 2]
            # >>> tokenizer("welcomes")['input_ids']
            # [0, 605, 523, 32696, 2]
            # >>> tokenizer("welcomes you")['input_ids']
            # [0, 605, 523, 32696, 47, 2]
            # -> so we skip this check here
            pass
        else:
            assert (
                len_text_id_with_special_tokens == calculated_len
            ), f"{len_text_id_with_special_tokens} == {calculated_len}"
        assert (
            count_nonzero_elements_in_target_mask_seq
            == num_effective_length_target_in_text
        )

        return target_mask_seq

    def _create_knowledge_source_tensor(
        self, text_tokens_as_str: List, offset: int, mapping, mode: str
    ):
        if mode == "nrc_emotions":
            num_categories = get_num_nrc_emotions()
        elif mode == "mpqa_subjectivity":
            num_categories = get_num_mpqa_subjectivity_polarities()
        elif mode == "bingliu_opinion":
            num_categories = get_num_bingliu_polarities()
        elif mode == "liwc":
            from knowledge.liwc.liwc import get_num_liwc_categories

            num_categories = get_num_liwc_categories()
        elif mode == "zeros":
            num_categories = get_num_zero_dimensions()
        else:
            raise NotImplementedError

        emotions_for_sequence = torch.zeros(
            self.max_seq_len, num_categories, dtype=torch.long
        )
        for word_index, word in enumerate(text_tokens_as_str, start=offset):
            positions_of_current_word_in_input_sequence = self._map_token_index_to_wordpiece_index(
                mapping, word_index
            )
            if mode == "nrc_emotions":
                emotion_tensor = get_nrc_emotions_as_tensor(word)
            elif mode == "mpqa_subjectivity":
                emotion_tensor = get_mpqa_subjectivity_polarities_as_tensor(word)
            elif mode == "bingliu_opinion":
                emotion_tensor = get_bingliu_polarities_as_tensor(word)
            elif mode == "liwc":
                from knowledge.liwc.liwc import get_liwc_categories_as_tensor

                emotion_tensor = get_liwc_categories_as_tensor(word)
            elif mode == "zeros":
                emotion_tensor = get_zeros_as_tensor(word)
            else:
                raise NotImplementedError

            for position_in_seq in positions_of_current_word_in_input_sequence:
                emotions_for_sequence[position_in_seq, :] = emotion_tensor
        return emotions_for_sequence

    def _create_mapping_from_tokenbased_to_wordpiece_based(self, text, tok_obj):
        offset = 10
        # for spacy, we need to remove leading spaces as they will yield a single
        # token
        text_without_leading_space = text.strip()
        nlp_text = nlp(text_without_leading_space)

        # get only non-single-space tokens, see
        # https://github.com/explosion/spaCy/issues/1707
        text_tokens = [token for token in nlp_text if not token.is_space]
        text_tokens_as_str = [token.text for token in nlp_text if not token.is_space]
        # in case there is a word at position k that is longer than the max seq len
        # when converted to wordpiece indexes, _create_word_to_wordpiece_mapping
        # returns all words up to including k-1
        mapping, text_tokens_as_str = self._create_word_to_wordpiece_mapping(
            tok_obj,
            text_tokens_as_str,
            for_text_with_special_tokens=True,
            offset=offset,
        )
        # ..., correspondingly, truncate text_tokens to same length
        text_tokens = text_tokens[: len(text_tokens_as_str)]
        return mapping, offset, text_tokens_as_str, text_tokens

    def _convert_non_null_to_one(self, lst):
        lst_one = []
        for item in lst:
            if item == 0:
                lst_one.append(0)
            else:
                lst_one.append(1)
        return lst_one

    def _map_token_index_to_wordpiece_index(self, a, b):
        is_equal_tensor = a == b
        is_equal_tensor = torch.nonzero(is_equal_tensor, as_tuple=False)
        is_equal_tensor = is_equal_tensor.flatten()
        return is_equal_tensor.tolist()

    def _create_dependency_tensor(
        self,
        text_tokens,
        text_tokens_as_str,
        mapping_token2wordpiece,
        mapping_token2wordpiece_offset,
    ):
        # create the dependency matrix
        dependency_tensor_of_tokens = self._calculate_dep_matrix(
            text_tokens, text_tokens_as_str
        )
        # create empty wordpiece based tensors
        dependency_tensor_of_wordpieces = torch.zeros(
            self.max_seq_len, self.max_seq_len, dtype=torch.float
        )

        for word_index_withoffset, word in enumerate(
            text_tokens_as_str, start=mapping_token2wordpiece_offset
        ):
            word_index = word_index_withoffset - mapping_token2wordpiece_offset
            dependency_column = dependency_tensor_of_tokens[word_index].tolist()
            dependency_column_ones = self._convert_non_null_to_one(dependency_column)
            _check_num_heads = sum(dependency_column_ones)
            assert _check_num_heads in [0, 1,], (
                f"expected only zero or one heads, found {_check_num_heads} for word "
                f"{word} in {text_tokens_as_str}"
            )
            if _check_num_heads == 0:
                # this can happen if the head is in the part of the sentence that was
                # truncated due to max seq len
                logger.warning(
                    "head of word %s is not part of the wordpiece sequence", word
                )
                continue

            head_index = dependency_column_ones.index(1)
            head_index_withoffset = head_index + mapping_token2wordpiece_offset
            dependency_type = dependency_column[head_index]
            assert dependency_type != 0

            positions_of_current_word_in_input_sequence = self._map_token_index_to_wordpiece_index(
                mapping_token2wordpiece, word_index_withoffset
            )
            positions_of_current_head_in_input_sequence = self._map_token_index_to_wordpiece_index(
                mapping_token2wordpiece, head_index_withoffset
            )

            for position_word in positions_of_current_word_in_input_sequence:
                for position_head in positions_of_current_head_in_input_sequence:
                    dependency_tensor_of_wordpieces[
                        position_word, position_head
                    ] = dependency_type

        return dependency_tensor_of_wordpieces

    def _create_dependency_tree_hop_distances_of_tokens_to_target(
        self,
        text_tokens,
        text_left_len,
        target_phrase,
        text_tokens_as_str,
        mapping_token2wordpiece,
        mapping_token2wordpiece_offset,
    ):
        """
        finds for each token in the sentence, or rather each token's wordpiece(s) the distance in the
        syntax tree to the target phrase. returns a list of size k where k is the maxseqlen; the list
        contains the distances for each wordpiece.
        :param text_tokens:
        :param text_left_len:
        :param target_phrase:
        :param text_tokens_as_str:
        :param mapping_token2wordpiece:
        :param mapping_token2wordpiece_offset:
        :return:
        """
        # Find distance in dependency parsing tree
        num_chars_counted = sum([len(token) for token in text_tokens_as_str])
        dist = self._calculate_dep_distance(text_tokens, text_left_len, target_phrase)

        # create empty wordpiece based tensors
        syntax_hop_distance_to_target = torch.zeros(self.max_seq_len, dtype=torch.long)

        # convert:
        # syntax_hop_distance_to_target and dependency_tensor_of_wordpieces is wordpiece
        # based. dist and dependency_tensor_of_tokens is token based. map from each to
        # the wordpiece variant
        for word_index, word in enumerate(
            text_tokens_as_str, start=mapping_token2wordpiece_offset
        ):
            word_index_without_offset = word_index - mapping_token2wordpiece_offset

            positions_of_current_word_in_input_sequence = self._map_token_index_to_wordpiece_index(
                mapping_token2wordpiece, word_index
            )

            depdistance_of_current_word_in_input_sequence = dist[
                word_index_without_offset
            ]

            for position in positions_of_current_word_in_input_sequence:
                syntax_hop_distance_to_target[
                    position
                ] = depdistance_of_current_word_in_input_sequence

        return syntax_hop_distance_to_target

    def create_model_input_seqs(
        self,
        text_left,
        target_phrase,
        text_right,
        coreferential_targets_for_target_mask: Iterable[dict],
    ):
        """
        Creates input sequences for a given target. Prior components have processed the jsonl
        so that coreferential_targets_for_target_mask will contain a list of coreferential
        mentions of the target if and only if coref_mode == "in_targetmask". In that case,
        will produce an individual target mask for each coref mention and merge them to
        the target mask of the preferred mention.
        :param text_left:
        :param target_phrase:
        :param text_right:
        :param further_mentions:
        :return:
        """
        text = self.create_entire_text(
            text_left, target_phrase, text_right, is_return_modified_text_left=False
        )
        logger.debug(f"'{text_left}' '{target_phrase}' '{text_right}'")

        input_seqs_per_tokenizer = {}

        for name, tok_obj in self.tokenizers_name_and_obj.items():
            logger.debug(f"{name}")

            # text
            text_ids_with_special_tokens = tok_obj.encode(
                text,
                max_length=self.max_seq_len,
                pad_to_max_length=True,
                add_special_tokens=True,
                truncation="longest_first",
            )
            # text SEP target
            adjusted_target_phrase = target_phrase
            if self.is_use_natural_target_phrase_for_spc:
                adjusted_target_phrase = "What do you think of " + target_phrase + "?"
            text_then_target_ids_with_special_tokens_dict = tok_obj.encode_plus(
                text,
                text_pair=adjusted_target_phrase,
                max_length=self.max_seq_len,
                pad_to_max_length=True,
                add_special_tokens=True,
                truncation="longest_first",
            )
            text_then_target_ids_with_special_tokens_dict = (
                text_then_target_ids_with_special_tokens_dict.data
            )
            text_then_target_ids_with_special_tokens = text_then_target_ids_with_special_tokens_dict[
                "input_ids"
            ]
            if type(tok_obj) == RobertaTokenizer:
                # roberta doesnt have segments, so we produce fake segment ids here, and later simply dont pass them
                text_then_target_ids_with_special_tokens_segment_ids = [0] * len(
                    text_then_target_ids_with_special_tokens
                )
            else:
                text_then_target_ids_with_special_tokens_segment_ids = text_then_target_ids_with_special_tokens_dict[
                    "token_type_ids"
                ]
            # target
            target_ids_with_special_tokens = tok_obj.encode(
                target_phrase,
                max_length=self.max_seq_len,
                pad_to_max_length=True,
                add_special_tokens=True,
                truncation="longest_first",
            )

            # create target masks
            target_mask_seq_for_text_with_special_tokens = self._create_target_mask(
                tok_obj,
                text_left,
                target_phrase,
                text_right,
                for_text_with_special_tokens=True,
            )
            # create target mask for coreferential targets of given target
            coref_target_masks = self._create_coreferential_target_masks(
                tok_obj, coreferential_targets_for_target_mask
            )
            # merge them into the preferred target mask
            merged_target_mask = self._merge_coref_target_masks_into_preferred_target_mask(
                target_mask_seq_for_text_with_special_tokens, coref_target_masks
            )
            target_mask_seq_for_text_with_special_tokens = merged_target_mask

            # create also text ids without max length to see if the one that will be
            # used was truncated
            text_ids_with_special_tokens_no_max_length = tok_obj.encode(
                text, add_special_tokens=True
            )
            text_num_truncated_tokens = len(
                text_ids_with_special_tokens_no_max_length
            ) - len(text_ids_with_special_tokens)

            # iterate tokens of full text and look up in dicts
            # create mapping from token-based indexes to wordpiece-based indexes
            # the function will also remove any tokens for which the word piece index
            # would be after max seq len
            # under certain circumstances, e.g., when using albert as well as the target
            # phrase is close to what would be cut off (but is not at this point), it can happen
            # that at a later point in time, i.e., _create_dependency_tree_hop_distances_of_tokens_to_target
            # or more specifically _calculate_dep_distance, the target cannot be found in the list of
            # tokens anymore, because this list of now cut off. thus, i changed the assert in
            # _calculate_dep_distance to throwing an exception that is catched later on
            (
                mapping_token2wordpiece,
                mapping_token2wordpiece_offset,
                text_tokens_as_str,
                text_tokens,
            ) = self._create_mapping_from_tokenbased_to_wordpiece_based(text, tok_obj)

            # create additional knowledge source tensors
            # stack only those knowledge sources that were requested by arguments
            selected_tensors_knowledge_sources_text = []
            for source in self.knowledge_sources:
                text_tensor = self._create_knowledge_source_tensor(
                    text_tokens_as_str,
                    mapping_token2wordpiece_offset,
                    mapping_token2wordpiece,
                    source,
                )
                selected_tensors_knowledge_sources_text.append(text_tensor)
            if self.knowledge_sources:
                text_stacked_knowledge_source_info = torch.cat(
                    tuple(selected_tensors_knowledge_sources_text), dim=1
                )
            else:
                text_stacked_knowledge_source_info = None

            # syntax hop distance
            text_dependency_tree_hop_distances = self._create_dependency_tree_hop_distances_of_tokens_to_target(
                text_tokens,
                len(text_left),
                target_phrase,
                text_tokens_as_str,
                mapping_token2wordpiece,
                mapping_token2wordpiece_offset,
            )

            text_dependency_matrix = self._create_dependency_tensor(
                text_tokens,
                text_tokens_as_str,
                mapping_token2wordpiece,
                mapping_token2wordpiece_offset,
            )

            # create item with indexes and masks
            input_seqs_per_tokenizer[name] = {
                FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS: torch.LongTensor(
                    text_ids_with_special_tokens
                ),
                FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK: torch.FloatTensor(
                    target_mask_seq_for_text_with_special_tokens
                ),
                FIELD_IS_OVERFLOW: text_num_truncated_tokens > 0,
                FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS: torch.LongTensor(
                    text_then_target_ids_with_special_tokens
                ),
                # we used to have text-then-target target mask here, but won't use it,
                # since it would be identical to the text target mask (since we only
                # want to mark the target within the text, but not in the 2nd target
                # component)
                # FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK: torch.LongTensor(
                #    text_then_target_ids_with_special_tokens_target_mask,
                # ),
                FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_SEGMENT_IDS: torch.LongTensor(
                    text_then_target_ids_with_special_tokens_segment_ids
                ),
                FIELD_TARGET_IDS_WITH_SPECIAL_TOKENS: torch.LongTensor(
                    target_ids_with_special_tokens
                ),
                FIELD_SYNTAX_HOP_DISTANCE_TO_TARGET: text_dependency_tree_hop_distances,
                FIELD_SYNTAX_DEPENDENCY_MATRIX: text_dependency_matrix,
            }

            # add knowledge source if requested
            if self.knowledge_sources:
                input_seqs_per_tokenizer[name][
                    FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_SELECTED_KNOWLEDGE_SOURCES
                ] = text_stacked_knowledge_source_info
                # likewise as for FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK, we
                # don't need to set a special knowledge source mask for text-then-target

        return input_seqs_per_tokenizer

    def _create_coreferential_target_masks(
        self, tok_obj, coreferential_targets_for_target_mask
    ):
        target_masks = []
        for coref_target in coreferential_targets_for_target_mask:
            coref_target_mask = self._create_target_mask(
                tok_obj,
                self.prepare_left_segment(coref_target["text_left"]),
                self.prepare_target_mention(coref_target["mention"]),
                "",
                for_text_with_special_tokens=True,
            )
            target_masks.append(coref_target_mask)
        return target_masks

    def _merge_coref_target_masks_into_preferred_target_mask(
        self, target_mask_seq_for_text_with_special_tokens, coref_target_masks
    ):
        if coref_target_masks is None or len(coref_target_masks) == 0:
            return target_mask_seq_for_text_with_special_tokens

        # target_mask_seq_for_text_with_special_tokens is a list of scalars, coref tms a list of lists
        # turn both into a list of lists and append them
        all_target_masks = [
            target_mask_seq_for_text_with_special_tokens
        ] + coref_target_masks
        assert len(all_target_masks) == 1 + len(coref_target_masks)
        num_scalars = len(target_mask_seq_for_text_with_special_tokens)
        # ensure all same length
        for lst in all_target_masks:
            assert len(lst) == num_scalars
        # create merged lst
        merged_target_mask = [0] * num_scalars
        for index in range(num_scalars):
            merged_value = 0
            for lst in all_target_masks:
                lst_cur_value = lst[index]
                assert lst_cur_value in [0, 1]
                merged_value += lst_cur_value
            merged_value = min(1, merged_value)
            merged_target_mask[index] = merged_value

        return merged_target_mask


class FXDataset(Dataset):
    __PARAMS2INDEX = {}
    NUM_MAX_TARGETS_PER_ITEM = 2
    SINGLE_TARGETS = False

    def __init__(
        self,
        filepath,
        data_format: str,
        tokenizer: FXEasyTokenizer,
        named_polarity_to_class_number,
        class_number_to_named_polarity,
        sorted_expected_label_names,
        single_targets: bool,
        coref_mode: str,
        devmode=False,
        ignore_parsing_errors=False,
    ):
        self.class_label2class_number = named_polarity_to_class_number
        self.class_number2class_label = class_number_to_named_polarity
        self.sorted_expected_label_names = sorted_expected_label_names
        self.tokenizer = tokenizer
        self.data = []
        self.data_format = data_format
        self.count_sequence_overflow = Counter()
        FXDataset.SINGLE_TARGETS = single_targets

        if FXDataset.SINGLE_TARGETS:
            # single mode is enabled, so there will be only 1 target per item
            FXDataset.NUM_MAX_TARGETS_PER_ITEM = 1

        logger.info("reading dataset file {}".format(filepath))

        tasks = []
        with jsonlines.open(filepath, "r") as reader:
            for task in reader:
                task["sentence_normalized"] = self.__prepare_text_field(
                    task["sentence_normalized"]
                )
                for target in task["targets"]:
                    target["mention"] = self.__prepare_text_field(target["mention"])
                    further_mentions = []
                    for further_mention in target.get("further_mentions", []):
                        further_mention["mention"] = self.__prepare_text_field(
                            further_mention["mention"]
                        )
                        further_mentions.append(further_mention)
                    target["further_mentions"] = further_mentions
                tasks.append(task)

        if devmode:
            k = 4
            logger.warning("DEV MODE IS ENABLED")
            logger.info("devmode=True: truncating dataset to {} lines".format(k))
            tasks = tasks[:k]

        self.label_counter = Counter()
        with tqdm(total=len(tasks)) as pbar:
            for task in tasks:
                # create model items from this jsonl row (in case of single_targets
                # enabled, a row with k targets will return k model items)
                model_items = self.task_to_dataset_item(
                    task, coref_mode, ignore_parsing_errors
                )
                for converted_item, count_labels in model_items:
                    if converted_item is None:
                        # this will happen if an example has 0 targets -> ignore
                        self.label_counter["ignored_example"] += 1
                    else:
                        self.label_counter += count_labels
                        self.data.append(converted_item)
                pbar.update(1)
        logger.info("label distribution: {}".format(self.label_counter))

    def __prepare_text_field(self, text: str):
        """
        Replaces "invalid" characters with something else of the same length (same
        length is important so that the offsets in the dataset are still valid)
        :param text:
        :return:
        """
        return text.replace("\xa0", " ").replace("��", "  ")

    def __create_target_text_components(
        self,
        target_start_char,
        target_end_char,
        text,
        check_mention,
        ignore_parsing_errors,
    ):
        target_check_mention = text[target_start_char:target_end_char]
        text_left = text[:target_start_char]
        text_right = text[target_end_char:]
        # perform target mention check
        if ignore_parsing_errors:
            if check_mention != target_check_mention:
                logger.warning(
                    f"indexes do not match: '{check_mention}' vs '{target_check_mention}' "
                    f"({text_left}; {text_right}; {target_start_char}; {target_end_char})"
                )
        else:
            assert check_mention == target_check_mention, (
                f"indexes do not match: '{check_mention}' vs '{target_check_mention}' "
                f"({text_left}; {text_right}; {target_start_char}; {target_end_char})"
            )

        return text_left, text_right

    def __parse_dataset_row(self, task, ignore_parsing_errors):
        if self.data_format == "newsmtsc":
            _text = task["sentence_normalized"]
            primary_mtsc_gid = task["primary_gid"]
            if not FXDataset.SINGLE_TARGETS:
                # if single_targets mode is not enabled, ensure each item has at most
                # FXDataset.NUM_MAX_TARGETS_PER_ITEM targets
                # assert 1 <= len(task["targets"]) <= FXDataset.NUM_MAX_TARGETS_PER_ITEM
                pass
            else:
                # if single_targets mode is enabled, FXDataset.NUM_MAX_TARGETS_PER_ITEM
                # was set to 1, so we cannot perform the check at this point. and also
                # do not need to ensure because later we will expand the items so that
                # resulting items each will only have 1 target
                pass

            # perform validity check of target mention
            # also, add text_left and text_right to each target
            for target in task["targets"]:
                target_ae_gid = target["Input.gid"]
                target_start_char = target["from"]
                target_end_char = target["to"]
                target_mention = target["mention"]
                target_polarity = target["polarity"]

                text_left, text_right = self.__create_target_text_components(
                    target_start_char,
                    target_end_char,
                    _text,
                    target_mention,
                    ignore_parsing_errors,
                )

                # add target-specific text_left and text_right
                target["text_left"] = text_left
                target["text_right"] = text_right

            targets = task["targets"]
            return primary_mtsc_gid, _text, targets
        else:
            raise ValueError(f"data format unknown: {self.data_format}")

    def _create_target_inputs(
        self,
        text_left: str,
        target_mention: str,
        text_right: str,
        target_gid: str,
        polarity: float,
        count_labels: Counter,
        coreferential_targets_for_target_mask: Iterable[dict],
        is_fillup: bool,
    ):
        # always convert polarity to float
        polarity = float(polarity)

        # earlier, we assumed basic tokanization, e.g., that "Mr. Smith's statement
        # is stupid." was already tokenized into "Mr. Smith 's statement is stupid ."
        # however, since both encode and encode_plus internally perform tokenization
        # when invoked with a str object, we do not need to assume that the input is
        # already tokenized.
        # here is how we handle the input text and target phrase
        # 1) we ensure that there is a whitespace between left and target. if there
        #    is no whitespace, we add a whitespace to the end of left
        # 2) we do the same as in (1) for target and right, and if necessary add a
        #    whitespace to the beginning of right
        # by doing these two operations, we ensure that any target-dependent mapping
        # e.g., the target mask, can be mapped 1:1 to the the input text. for
        # example, "<target>Mr. Smith</target>'s statement is stupid." will be
        # converted into "Mr. Smith 's statement ..." this way, the LM's encode or
        # encode_plus function will (after internal tokenization) yield a list of
        # input indexes that we can unambiguously map to the target mask that we
        # produce ourselves, e.g., in FXEasyTokenizer:_create_target_mask. said
        # differently, by adding the whitespace, we ensure that "Mr. Smith" will be
        # mapped to its own indexes, whereas if we passed "Mr. Smith's", this might
        # yield different indexes, e.g., if "Smith's" is part of the LM's
        # tokenizer's vocabulary. for more information, see NA-633 and NA-643.
        text_left = FXEasyTokenizer.prepare_left_segment(text_left)
        target_mention = FXEasyTokenizer.prepare_target_mention(target_mention)
        text_right = FXEasyTokenizer.prepare_right_segment(text_right)
        text = FXEasyTokenizer.create_entire_text(
            text_left, target_mention, text_right, is_return_modified_text_left=False,
        )

        # text to indexes
        try:
            inputs_of_target = self.tokenizer.create_model_input_seqs(
                text_left,
                target_mention,
                text_right,
                coreferential_targets_for_target_mask,
            )
        except TooLongTextException:
            logger.error(
                "text is too long for target mask creation for target_gid=%s, ignoring target",
                target_gid,
            )
            return None
        except TargetNotFoundException as tnfe:
            logger.error(
                "target was not found after preprocessing when calculating "
                "distance hop for target_gid=%s, ignoring target (error msg: %s)",
                target_gid,
                str(tnfe),
            )
            return None

        # count overflow
        for tokenizer_name in self.tokenizer.tokenizers_name_and_obj.keys():
            if inputs_of_target[tokenizer_name]["is_overflow"]:
                self.count_sequence_overflow[tokenizer_name] += 1
                self.count_sequence_overflow["all"] += 1

        # add polarity
        inputs_of_target["polarity"] = SentimentClasses.polarity2normalized_polarity(
            polarity
        )
        inputs_of_target["label"] = SentimentClasses.polarity2label(polarity)
        # update stats
        count_labels[inputs_of_target["label"]] += 1

        # add reference information
        inputs_of_target["target_gid"] = target_gid

        # add original text and target (we can use that to create a mistake table)
        inputs_of_target["orig_text"] = text
        inputs_of_target["orig_target"] = target_mention
        inputs_of_target["orig_polarity"] = polarity

        # add whether fill up target (fake) or real
        inputs_of_target["is_fillup"] = is_fillup

        return inputs_of_target

    def _create_fillup_target(self):
        return self._create_target_inputs(
            text_left=" ",
            target_mention="fillup",
            text_right=" ",
            target_gid="fillup",
            polarity=SentimentClasses.FILLUP_POLARITY_VALUE,
            count_labels=Counter(),
            coreferential_targets_for_target_mask=[],
            is_fillup=True,
        )

    def _get_as_list(self, lst_inputs_of_targets: List[dict], key: str):
        """
        selects the value under the key in each
        list element of lst_inputs_of_targets. returns these values as a new list
        :param lst_inputs_of_targets:
        :param key:
        :return:
        """
        new_list = []
        for inputs_of_targets in lst_inputs_of_targets:
            new_list.append(inputs_of_targets[key])
        return new_list

    def _get_values_as_lists_in_dict(
        self, lst_inputs_of_targets: List[dict], keys: Iterable[str]
    ):
        d = {}
        for key in keys:
            d[key] = self._get_as_list(lst_inputs_of_targets, key)
        return d

    def _convert_multi_targets_in_single_item_to_single_target_in_multi_items(
        self, example_id, text, targets
    ):
        """
        Converts a single item with k targets into k items with 1 target each
        :param example_id:
        :param text:
        :param targets:
        :return: list of tuples (example_id, text, targets)
        """
        single_items = []
        for target in targets:
            single_example_id = target["Input.gid"]
            single_text = text
            single_targets = [target]
            single_item = (single_example_id, single_text, single_targets)
            single_items.append(single_item)
        return single_items

    def _convert_multi_targets_in_single_item_to_k_targets_in_multi_items(
        self, example_id, text, targets
    ):
        """
        Converts a single item with m targets into ceil(m/k) items with k items. If a
        resulting item would have less than k targets, its existing targets are repeated
        so that each item will have exactly k targets (in this case with duplicate
        targets).
        :param example_id:
        :param text:
        :param targets:
        :return:
        """
        shuffled_targets = targets.copy()
        random.shuffle(shuffled_targets)
        num_targets_per_item = FXDataset.NUM_MAX_TARGETS_PER_ITEM
        logger.debug(shuffled_targets)

        items = []
        count_processed_targets = 0
        for i in range(0, len(shuffled_targets), num_targets_per_item):
            cur_targets = shuffled_targets[i : i + num_targets_per_item]
            if len(cur_targets) < num_targets_per_item:
                logger.debug("too few targets")
                break
            assert len(cur_targets) == num_targets_per_item
            virtual_example_id = example_id + "_" + str(i)
            items.append((virtual_example_id, text, cur_targets))
            count_processed_targets += len(cur_targets)

        # process remaining targets
        logger.debug("processing remaining targets")
        if count_processed_targets < len(shuffled_targets):
            remaining_targets = shuffled_targets[count_processed_targets:]
            assert len(remaining_targets) < num_targets_per_item

            # duplicate targets until we have enough targets
            # create intermediate list to not skew random selection because of adding
            # the duplicates immediately
            tmp_targets = []
            while len(remaining_targets) + len(tmp_targets) < num_targets_per_item:
                tmp_targets.append(random.choice(remaining_targets))
                logger.debug(
                    f"{len(remaining_targets)}, {len(tmp_targets)}, {num_targets_per_item}"
                )
            remaining_targets.extend(tmp_targets)
            assert len(remaining_targets) == num_targets_per_item
            virtual_example_id = example_id + "_last"
            items.append((virtual_example_id, text, remaining_targets))

        # validation
        count_targets = 0
        seen_targets = []
        for item in items:
            targets = item[2]
            count_targets += len(targets)
            for target in targets:
                seen_targets.append(target)
        assert count_targets % num_targets_per_item == 0
        # seen targets must either be of same length as shuffled target, or - in case of
        # added duplicate targets - larger
        assert len(seen_targets) >= len(
            shuffled_targets
        ), f"{len(seen_targets)} vs. {len(shuffled_targets)}"

        return items

    def _create_item_for_model(self, example_id, text, targets):
        lst_inputs_of_targets = []
        count_labels = Counter()

        for target in targets:
            text_left = target["text_left"]
            text_right = target["text_right"]
            target_mention = target["mention"]
            polarity = target["polarity"]
            target_gid = target["Input.gid"]
            coreferential_targets_for_target_mask = target.get(
                "coreferential_targets_for_target_mask", []
            )

            inputs_of_target = self._create_target_inputs(
                text_left=text_left,
                target_mention=target_mention,
                text_right=text_right,
                target_gid=target_gid,
                polarity=polarity,
                count_labels=count_labels,
                coreferential_targets_for_target_mask=coreferential_targets_for_target_mask,
                is_fillup=False,
            )
            if inputs_of_target is None:
                # this will happen if a target mask could not be created because the
                # text_left is too long. -> ignore this target
                pass
            else:
                lst_inputs_of_targets.append(inputs_of_target)
        num_actual_targets = len(lst_inputs_of_targets)

        if num_actual_targets == 0:
            logger.error(
                "ignoring example (gid=%s) because it has 0 targets", example_id
            )
            return None, None

        # due to torch's requirements in dataloader, we need fill up remaining targets
        # with fake targets
        while len(lst_inputs_of_targets) < FXDataset.NUM_MAX_TARGETS_PER_ITEM:
            lst_inputs_of_targets.append(self._create_fillup_target())

        # lastly, merge inputs to one tensor over all targets
        # this also removes some fields, e.g., list of strs. should we need the removed
        # information later, we could simply add a new field below to "converted_item"
        stacked_targets, skipped_keys = self._stack_target_inputs(lst_inputs_of_targets)

        # create item dict
        # originally, we also had original targets here, but they get scrambled up and
        # there is no StrTensor to easily preserve their order
        converted_item = {
            "text": text,
            "num_targets": num_actual_targets,
            "primary_gid": example_id,
        }

        # merge stacked_targets dict into item
        # first, though, ensure no overlap by key
        assert len(set(converted_item.keys()).intersection(stacked_targets.keys())) == 0
        # now merge
        converted_item = {**converted_item, **stacked_targets}

        return converted_item, count_labels

    def create_virtual_target_from_target_mention(self, target, mention, text):
        mention_from = mention["from"]
        mention_to = mention["to"]
        mention_text = mention["mention"]

        mention_gid = target["Input.gid"] + "__" + mention_text
        mention_polarity = target["polarity"]

        text_left, text_right = self.__create_target_text_components(
            mention_from, mention_to, text, mention_text, False
        )

        return {
            "Input.gid": mention_gid,
            "from": mention_from,
            "to": mention_to,
            "mention": mention_text,
            "polarity": mention_polarity,
            "text_left": text_left,
            "text_right": text_right,
        }

    def _expand_coref_mentions_of_targets_to_multiple_targets(
        self, targets: Iterable, text: str
    ):
        expanded_targets = []
        for target in targets:
            expanded_targets.append(target)

            further_mentions = target.get("further_mentions", [])
            for further_mention in further_mentions:
                virtual_target = self.create_virtual_target_from_target_mention(
                    target, further_mention, text
                )
                expanded_targets.append(virtual_target)

        return expanded_targets

    def task_to_dataset_item(self, task, coref_mode: str, ignore_parsing_errors):
        # parse current dataset row (=task)
        example_id, text, targets = self.__parse_dataset_row(
            task, ignore_parsing_errors
        )

        # if the coref mode (only during training though) is additional_examples, repeat them here
        if coref_mode == "additional_examples":
            expanded_targets = self._expand_coref_mentions_of_targets_to_multiple_targets(
                targets, text
            )
        elif coref_mode == "in_targetmask":
            expanded_targets = targets
            for target in expanded_targets:
                further_mentions = target.get("further_mentions", [])
                targets_for_target_mask = []
                for further_mention in further_mentions:
                    virtual_target = self.create_virtual_target_from_target_mention(
                        target, further_mention, text
                    )
                    targets_for_target_mask.append(virtual_target)
                target[
                    "coreferential_targets_for_target_mask"
                ] = targets_for_target_mask
        else:
            # if coref mode is ignore, nothing to do here
            expanded_targets = targets

        # if single_targets mode is enabled, we want to have only one target for each
        # sentence, thus expand sentences
        if FXDataset.SINGLE_TARGETS:
            items = self._convert_multi_targets_in_single_item_to_single_target_in_multi_items(
                example_id, text, expanded_targets
            )
        else:
            items = self._convert_multi_targets_in_single_item_to_k_targets_in_multi_items(
                example_id, text, expanded_targets
            )

        # iterate (virtual) items of this actual row in the jsonl and create model items
        model_items = []
        for example_id, text, targets in items:
            res = self._create_item_for_model(example_id, text, targets)
            model_items.append(res)

        # return
        return model_items

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return len(self.data)

    @staticmethod
    def set_params2index(params2index: dict):
        if len(FXDataset.__PARAMS2INDEX) > 0:
            # verify the new one is same
            assert FXDataset.__PARAMS2INDEX == params2index
        else:
            FXDataset.__PARAMS2INDEX = params2index

    @staticmethod
    def get_input_by_params(inputs: List, weight_name: str, field_name: str):
        """
        If FXDataset.SINGLE_TARGETS is True, include the target dimension (currently 2nd
        dimension in tensor (dims are: batch, target, indexes or other values). If
        False, remove the target dimension.
        :param inputs:
        :param weight_name:
        :param field_name:
        :param include_target_dimension:
        :return:
        """
        param_id = (weight_name, field_name)
        seq = inputs[FXDataset.__PARAMS2INDEX[param_id]]
        if not FXDataset.SINGLE_TARGETS:
            return seq
        else:
            assert seq.shape[1] == 1, "there are more than one targets"
            return seq.squeeze(dim=1)

    @staticmethod
    def get_all_inputs_for_model(inputs: List, model_class: FXBaseModel):
        selected_inputs = []
        for weight_name, field_name in model_class.get_input_field_ids():
            selected_inputs.append(
                FXDataset.get_input_by_params(inputs, weight_name, field_name)
            )
        return inputs

    def _stack_target_inputs(self, lst_inputs_of_targets: List[dict]) -> dict:
        """
        stacks each field to one Tensor
        :param lst_inputs_of_targets:
        :return:
        """
        # ensure correct number of targets
        assert len(lst_inputs_of_targets) == self.NUM_MAX_TARGETS_PER_ITEM
        # ensure each target has same keys
        all_keys = list(lst_inputs_of_targets[0].keys())
        for inputs_of_target in lst_inputs_of_targets:
            assert set(all_keys) == inputs_of_target.keys()
        # for each key, merge all targets
        merged_targets = {}

        # to collect keys that were skipped
        skipped_keys = []

        # keys that not "scalars", e.g., polarity, but dicts, e.g., "bert-base-uncased",
        # which itself holds multiple vectors, require special handling
        special_keys = list(self.tokenizer.tokenizers_name_and_obj.keys())
        # iterate keys
        for key in all_keys:
            if key in special_keys:
                merged_targets[key] = self._stack_special_field(
                    key, lst_inputs_of_targets
                )
            else:
                inputs_over_targets_of_key = []
                for inputs_of_target in lst_inputs_of_targets:
                    inputs_over_targets_of_key.append(inputs_of_target[key])

                # merge targets' inputs to one tensor
                stacked_inputs_over_targets_of_key = self._stack_or_create_tensor(
                    inputs_over_targets_of_key
                )
                # add stacked tensor to dict
                if stacked_inputs_over_targets_of_key is not None:
                    merged_targets[key] = stacked_inputs_over_targets_of_key
                else:
                    # if None, dont add
                    logger.debug("skipping key: %s", key)
                    skipped_keys.append(key)

        return merged_targets, skipped_keys

    def _stack_special_field(self, key: str, lst_inputs_of_targets: List[dict]):
        # ensure each target has same sub keys for current selected key
        all_sub_keys = list(lst_inputs_of_targets[0][key].keys())
        for inputs_of_target in lst_inputs_of_targets:
            assert set(all_sub_keys) == inputs_of_target[key].keys()

        # for each sub key, merge all targets
        merged_targets = {}

        for sub_key in all_sub_keys:
            inputs_over_targets_of_sub_key = []
            for inputs_of_target in lst_inputs_of_targets:
                # select
                inputs = inputs_of_target[key][sub_key]
                inputs_over_targets_of_sub_key.append(inputs)
            # merge targets' inputs to one tensor
            stacked_inputs_over_targets_of_sub_key = self._stack_or_create_tensor(
                inputs_over_targets_of_sub_key
            )
            # add stacked tensor to dict
            if stacked_inputs_over_targets_of_sub_key is not None:
                merged_targets[sub_key] = stacked_inputs_over_targets_of_sub_key
            else:
                # if None, dont add
                logger.debug("skipping key: %s", sub_key)

        return merged_targets

    def _stack_or_create_tensor(self, lst_values: List):
        """
        Given a list of values, checks if all are of same type and creates a single
        Tensor from the values (either by stacking or by creation)
        :param lst_values:
        :return:
        """
        # assert all same type
        first_type = type(lst_values[0])
        for val in lst_values:
            assert first_type == type(val), f"{first_type} vs {type(val)}"

        # stack list of tensors to one tensor
        if first_type == bool:
            # if bool, create a new tensor
            stacked_inputs_over_targets_of_sub_key = torch.BoolTensor(lst_values)
        elif first_type == int:
            # we need to have Long here, otherwise will run into an error during loss
            # calculation, where Long is expected
            stacked_inputs_over_targets_of_sub_key = torch.LongTensor(lst_values)
        elif first_type == float:
            stacked_inputs_over_targets_of_sub_key = torch.FloatTensor(lst_values)
        elif first_type == str:
            stacked_inputs_over_targets_of_sub_key = None
        elif torch.is_tensor(lst_values[0]):
            # if Tensor of subtype, stack
            stacked_inputs_over_targets_of_sub_key = torch.stack(lst_values)
        else:
            raise NotImplementedError(f"no merging for type: {first_type}")

        return stacked_inputs_over_targets_of_sub_key



from argparse import Namespace
from typing import List, Dict
from NewsSentiment.consts import *
import torch.nn as nn
import torch

from NewsSentiment.models.FXBaseModel import FXBaseModel
from NewsSentiment.models.singletarget.td_bert import TD_BERT


class EnsembleTopB(FXBaseModel):
    @staticmethod
    def get_language_models():
        return (get_default_lm(),)

    @staticmethod
    def get_input_field_ids():
        return [
            # tdbert
            (get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
            # hosseinia
            (get_default_lm(), FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS),
            (get_default_lm(), FIELD_TEXT_THEN_TARGET_IDS_WITH_SPECIAL_TOKENS_SEGMENT_IDS),
            # while we use text-then-target as bert input, we can use text targetmask and text knowledge source mask
            # because it is identical to a hypothetical text-then-target target mask or text-then-target knowledge
            # source mask (we would not highlight the target in the 2nd component in the corresponding mask)
            (get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
            (get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_SELECTED_KNOWLEDGE_SOURCES),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super().__init__()
        from models.singletarget.knowledgesourceshosseiniatdbert import KnowledgeSourcesHosseiniaTdBert

        # submodels models
        self.td_bert = TD_BERT(transformer_models, opt)
        self.hosseiniatdbert = KnowledgeSourcesHosseiniaTdBert(transformer_models, opt)

        # ensemble related
        self.ensemble_combiner = nn.Linear(
            self.hosseiniatdbert.language_model.config.hidden_size * 2,
            opt.polarities_dim
        )

    def forward(self, inputs: List):
        # shape: batch, bertdim
        td_bert_out = self.td_bert(inputs, is_return_ensemble_values=True)
        # shape: batch, 3 * 2 * 2 * bertdim
        hosseinia_out = self.hosseiniatdbert(inputs, is_return_ensemble_values=True)

        # combine
        combined_out = torch.cat((td_bert_out, hosseinia_out), dim=1)

        logits = self.ensemble_combiner(combined_out)

        return logits



# -*- coding: utf-8 -*-
# file: point_wise_feed_forward.py
# author: songyouwei <youwei0314@gmail.com>
# Copyright (C) 2018. All Rights Reserved.

import torch.nn as nn


class PositionwiseFeedForward(nn.Module):
    ''' A two-feed-forward-layer module '''
    def __init__(self, d_hid, d_inner_hid=None, dropout=0):
        super(PositionwiseFeedForward, self).__init__()
        if d_inner_hid is None:
            d_inner_hid = d_hid
        self.w_1 = nn.Conv1d(d_hid, d_inner_hid, 1)  # position-wise
        self.w_2 = nn.Conv1d(d_inner_hid, d_hid, 1)  # position-wise
        self.dropout = nn.Dropout(dropout)
        self.relu = nn.ReLU()

    def forward(self, x):
        output = self.relu(self.w_1(x.transpose(1, 2)))
        output = self.w_2(output).transpose(2, 1)
        output = self.dropout(output)
        return output



import argparse

from NewsSentiment.download import Download

if __name__ == '__main__':
    parser = argparse.ArgumentParser(prog='NewsSentiment')
    subparsers = parser.add_subparsers(dest='action')

    subparser_download = subparsers.add_parser('download', help=Download.add_subparser.__doc__)
    Download.add_subparser(subparser_download)

    args = parser.parse_args()
    action = args.action
    del args.action

    if action == 'download':
        Download.run_from_parser(args)



from typing import Iterable


class SentimentClasses:
    FILLUP_POLARITY_VALUE = -100
    FILLUP_POLARITY_LABEL = "fillup"
    SENTIMENT_CLASSES = None

    @staticmethod
    def initialize(sentiment_classes: dict):
        SentimentClasses.SENTIMENT_CLASSES = sentiment_classes

    @staticmethod
    def get_num_classes():
        return len(SentimentClasses.SENTIMENT_CLASSES)

    @staticmethod
    def __evaluate_boundary(given_value: float, boundary: tuple):
        operator = boundary[0]
        value = boundary[1]
        if operator == "<=":
            return given_value <= value
        elif operator == "<":
            return given_value < value
        elif operator == ">=":
            return given_value >= value
        elif operator == ">":
            return given_value > value
        elif operator == "==":
            return given_value == value
        else:
            raise ValueError

    @staticmethod
    def __evaluate_boundaries_of_class(
        given_value: float, sentiment_boundaries: Iterable[tuple]
    ):
        assert len(sentiment_boundaries) >= 1
        for boundary in sentiment_boundaries:
            is_valid = SentimentClasses.__evaluate_boundary(given_value, boundary)
            if not is_valid:
                return False
        return True

    @staticmethod
    def __get_legacy_information():
        # self.polarity_associations = {"positive": 2, "neutral": 1, "negative": 0}
        # self.polarity_associations_inv = {2: "positive", 1: "neutral", 0: "negative"}
        # self.sorted_expected_label_values = [0, 1, 2]
        # self.sorted_expected_label_names = ["negative", "neutral", "positive"]

        sentiment_labels = list(SentimentClasses.SENTIMENT_CLASSES.keys())
        sentiment_normalized_values = []
        for label in sentiment_labels:
            sentiment_normalized_values.append(
                SentimentClasses.SENTIMENT_CLASSES[label]["normalized_polarity"]
            )

        polarity_associations = {}
        polarity_associations_inv = {}
        for label, value in zip(sentiment_labels, sentiment_normalized_values):
            polarity_associations[label] = value
            polarity_associations_inv[value] = label

        return {
            "polarity_associations": polarity_associations,
            "polarity_associations_inv": polarity_associations_inv,
            "sorted_expected_label_values": sentiment_normalized_values,
            "sorted_expected_label_names": sentiment_labels,
        }

    @staticmethod
    def get_sorted_expected_label_names():
        return SentimentClasses.__get_legacy_information()[
            "sorted_expected_label_names"
        ]

    @staticmethod
    def get_sorted_expected_label_values():
        return SentimentClasses.__get_legacy_information()[
            "sorted_expected_label_values"
        ]

    @staticmethod
    def get_polarity_associations():
        return SentimentClasses.__get_legacy_information()["polarity_associations"]

    @staticmethod
    def get_polarity_associations_inverse():
        return SentimentClasses.__get_legacy_information()["polarity_associations_inv"]

    @staticmethod
    def __find_sentiment_class(polarity: float):
        resulting_class = None
        for sentiment_label, info in SentimentClasses.SENTIMENT_CLASSES.items():
            sentiment_boundaries = info["boundaries"]
            sentiment_normalized_polarity = info["normalized_polarity"]
            is_in_class_boundaries = SentimentClasses.__evaluate_boundaries_of_class(
                polarity, sentiment_boundaries
            )
            if is_in_class_boundaries:
                # check polarity is not in another class, too
                assert (
                    resulting_class is None
                ), f"overlapping sentiment classes; previous class: {resulting_class}"
                resulting_class = (sentiment_label, sentiment_normalized_polarity)

        # check that a class was found
        assert resulting_class, f"result is not defined for polarity: {polarity}"

        return resulting_class

    @staticmethod
    def polarity2label(polarity: float) -> str:
        if polarity == SentimentClasses.FILLUP_POLARITY_VALUE:
            return SentimentClasses.FILLUP_POLARITY_LABEL

        sentiment_class = SentimentClasses.__find_sentiment_class(polarity)
        label = sentiment_class[0]
        return label

    @staticmethod
    def polarity2normalized_polarity(polarity: float) -> int:
        if polarity == SentimentClasses.FILLUP_POLARITY_VALUE:
            return int(SentimentClasses.FILLUP_POLARITY_VALUE)

        sentiment_class = SentimentClasses.__find_sentiment_class(polarity)
        normalized_polarity = sentiment_class[1]
        return normalized_polarity

    @staticmethod
    def Sentiment3ForNewsMtsc():
        sentiment_classes = {
            "positive": {
                "boundaries": [(">=", 5), ("<=", 7)],
                "normalized_polarity": 2,
            },
            "neutral": {"boundaries": [(">", 3), ("<", 5)], "normalized_polarity": 1},
            "negative": {
                "boundaries": [(">=", 1), ("<=", 3)],
                "normalized_polarity": 0,
            },
        }
        SentimentClasses.initialize(sentiment_classes)

    @staticmethod
    def SentimentStrong3ForNewsMtsc():
        sentiment_classes = {
            "positive": {
                "boundaries": [(">=", 6), ("<=", 7)],
                "normalized_polarity": 2,
            },
            "neutral": {"boundaries": [(">", 2), ("<", 6)], "normalized_polarity": 1},
            "negative": {
                "boundaries": [(">=", 1), ("<=", 2)],
                "normalized_polarity": 0,
            },
        }
        SentimentClasses.initialize(sentiment_classes)

    @staticmethod
    def SentimentWeak3ForNewsMtsc():
        sentiment_classes = {
            "positive": {
                "boundaries": [(">=", 4.5), ("<=", 7)],
                "normalized_polarity": 2,
            },
            "neutral": {
                "boundaries": [(">", 3.5), ("<", 4.5)],
                "normalized_polarity": 1,
            },
            "negative": {
                "boundaries": [(">=", 1), ("<=", 3.5)],
                "normalized_polarity": 0,
            },
        }
        SentimentClasses.initialize(sentiment_classes)



from argparse import Namespace
from typing import Dict

import torch.nn as nn

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.layers.AggregatorForBert import AggregatorForBert
from NewsSentiment.models.FXBaseModel import FXBaseModel


class NoTargetClsBert(FXBaseModel):
    @staticmethod
    def get_language_models():
        return (BERT_BASE_UNCASED,)

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(NoTargetClsBert, self).__init__()
        self.language_model = transformer_models[BERT_BASE_UNCASED]
        self.aggregator_for_bert = AggregatorForBert(opt.spc_lm_representation)
        self.dropout = nn.Dropout(opt.dropout)
        self.dense = nn.Linear(
            self.language_model.config.hidden_size, opt.polarities_dim
        )

    def forward(self, inputs):
        text_bert_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )

        last_hidden_state, pooler_output, all_hidden_states = self.language_model(
            input_ids=text_bert_indices
        )
        prepared_output = self.aggregator_for_bert(
            last_hidden_state, pooler_output, all_hidden_states
        )

        prepared_output = self.dropout(prepared_output)
        logits = self.dense(prepared_output)

        return logits



# this file re-implements TD-BERT by Gao Zhengjie et al.
# while this file aims to be conceptually identical to TD-BERT, one technical difference is that we do not calculate
# the target mask within the model (here) but do this step as part of the dataset processing. in case there are strong
# performance differences between original TD-BERT and this implementation, this technical difference might be worth
# exploring whether it actually yields an identical implementation.
from argparse import Namespace
from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.models.FXBaseModel import FXBaseModel


class TD_BERT(FXBaseModel):
    @staticmethod
    def get_language_models():
        return (get_default_lm(),)

    @staticmethod
    def get_input_field_ids():
        return [
            (get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(TD_BERT, self).__init__()
        self.opt = opt
        self.language_model = transformer_models[get_default_lm()]
        self.dropout = nn.Dropout(opt.dropout)
        self.fc = nn.Linear(self.language_model.config.hidden_size, opt.polarities_dim)

    def forward(self, inputs, is_return_ensemble_values: bool = False):
        # get inputs
        text_bert_indices = FXDataset.get_input_by_params(
            inputs, get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        text_bert_indices_target_mask = FXDataset.get_input_by_params(
            inputs, get_default_lm(), FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK
        )

        # apply bert
        last_hidden_states = self.invoke_language_model(
            lm=self.language_model,
            input_ids=text_bert_indices,
        )

        # element-wise multiplication with target mask
        # unsqueeze, cf. https://stackoverflow.com/q/62559382
        text_bert_indices_target_mask_unsqueezed = text_bert_indices_target_mask.unsqueeze(
            -1
        )
        last_hidden_states_only_target = (
                last_hidden_states * text_bert_indices_target_mask_unsqueezed
        )

        # as in TD-BERT, perform max pooling
        last_hidden_states_only_target_aggregated, _ = last_hidden_states_only_target.max(
            dim=1
        )

        # dropout before dense layer, as in most other tsc models
        last_hidden_states_only_target_aggregated = self.dropout(
            last_hidden_states_only_target_aggregated
        )

        if is_return_ensemble_values:
            return last_hidden_states_only_target_aggregated
        else:
            # dense layer
            logits = self.fc(last_hidden_states_only_target_aggregated)
            # removed tanh, which was invoked in original tdbert. for training, we dont
            # need it to properly compute the loss. we would, however, need softmax during
            # inferring to have the probabilities of all mutually exclusive classes
            # to sum up to 1
            return logits






import torch
import torch.nn as nn

from NewsSentiment.SentimentClasses import SentimentClasses


class SequenceLoss(nn.Module):
    """
    Input to this loss are sequences, see models/multitargets/seq2seq.py
    """

    def __init__(self, device, weight):
        super(SequenceLoss, self).__init__()
        self.device = device
        self.weight = weight
        self.actual_loss = nn.CrossEntropyLoss(
            weight=self.weight, ignore_index=SentimentClasses.FILLUP_POLARITY_VALUE,
        )

        assert self.weight is None, "not implemented, weight must be None"

    def forward(
        self,
        predictions: torch.Tensor,
        true_classes: torch.Tensor,
        true_target_mask: torch.Tensor,
    ):
        """
        :param predictions:         shape: batch, numclasses, seqlen
        :param true_classes:        shape: batch, targets
        :param true_target_mask:    shape: batch, targets, seqlen
        :return:
        """
        batch_size = predictions.shape[0]
        num_classes = predictions.shape[1]
        seq_len = predictions.shape[2]

        # create true_sequence, shape: batch, seqlen
        true_sequence = (
            torch.ones([batch_size, seq_len], dtype=torch.int64, device=self.device)
            * SentimentClasses.FILLUP_POLARITY_VALUE
        )
        # update individual scalars
        for batch_index in range(batch_size):
            for target_index in range(true_target_mask.shape[1]):
                # get the sentiment class of this target in this batch
                true_sentiment_class = true_classes[batch_index, target_index].item()
                # if the true_sentiment_class is FILLUP_POLARITY_VALUE, ignore (no need
                # to update because we initialized the whole true_sequence with
                # FILLUP_POLARITY_VALUE
                if true_sentiment_class == SentimentClasses.FILLUP_POLARITY_VALUE:
                    continue

                # if not FILLUP_POLARITY_VALUE, update the true_sequence
                # iterate all tokens
                for seq_index in range(seq_len):
                    # determine if at the current token there is a target
                    is_target = true_target_mask[
                        batch_index, target_index, seq_index
                    ].item()

                    if is_target == 1:
                        # there is a target
                        # we should update the scalar in true_sequence at the
                        # corresponding part (thereby disregarding the target dimension,
                        # since all targets are merged into one dimension). ensure, that
                        # -100 is still there (if there is another value, this means
                        # that we have overlapping targets)
                        prev_value = true_sequence[batch_index, seq_index].item()
                        if prev_value == SentimentClasses.FILLUP_POLARITY_VALUE:
                            # the previous value is FILLUP_POLARITY_VALUE, so there is
                            # no target at this token already. so, we can update
                            true_sequence[batch_index, seq_index] = true_sentiment_class
                        else:
                            # there is already a target class -> overlapping targets
                            # this can happen for probably two reasons:
                            # 1) there are actually different targets in the data that
                            #    overlap
                            # 2) we duplicated a target in FXDataset
                            # either way, for now, if the value to be set is identical
                            # to the one already present, we continue, otherwise throw
                            # an error
                            if prev_value == true_sentiment_class:
                                pass
                            else:
                                raise ValueError(
                                    f"tried to update true_sequence[{batch_index},{seq_index}]={prev_value}"
                                )

                    elif is_target == 0:
                        # no target
                        # since we initialized the true_sequence tensor with -100
                        # scalars, there's not need to update the value for non-target
                        # nodes
                        pass
                    else:
                        raise ValueError(
                            f"true_target_mask must be either 0 or 1, is: {is_target}"
                        )

        loss = self.actual_loss(predictions, true_sequence)
        return loss



from argparse import Namespace
from typing import Dict

import torch
import torch.nn as nn

from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.models.FXBaseModel import FXBaseModel


class TDBertLikeSingle(FXBaseModel):
    """
    This model returns uses a target mask for a single target to obtain only the hidden
    states of those last layer nodes that correspond to a wordpiece of the target
    phrase. Then aggregation, dropout and a dense layer is applied to retrieve the
    3-class logits.

    From a conceptual perspective, TD-BERT is in some aspects similar (e.g., only the
    output of target-phrase-related nodes is used), but in other aspects not similar,
    e.g., we don't use max pooling but a mean on all non-0 target nodes. In other cases,
    it is not quite clear how the authors of TD-BERT implemented them. An email I sent
    to them was not answered yet.
    """

    @staticmethod
    def get_language_models():
        return (BERT_BASE_UNCASED,)

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(TDBertLikeSingle, self).__init__()
        self.language_model = transformer_models[BERT_BASE_UNCASED]
        self.dropout = nn.Dropout(opt.dropout)
        self.dense = nn.Linear(
            self.language_model.config.hidden_size, opt.polarities_dim
        )

    def forward(self, inputs):
        # get inputs
        text_bert_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )
        text_bert_indices_target_mask = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK
        )

        # apply bert
        last_hidden_states, pooler_output, all_hidden_states = self.language_model(
            input_ids=text_bert_indices
        )

        # element-wise multiplication with target mask
        # unsqueeze, cf. https://stackoverflow.com/q/62559382
        text_bert_indices_target_mask_unsqueezed = text_bert_indices_target_mask.unsqueeze(
            -1
        )
        last_hidden_states_only_target = (
            last_hidden_states * text_bert_indices_target_mask_unsqueezed
        )

        # similar to TD-BERT, perform max pooling TODO not implemented yet, instead:
        # for now, retrieve only the values of the target's output tokens and then
        # calculate the mean: (batchsize, 150, 768) -> (batchsize, 768)
        # get the positions of target nodes. note that we cannot simply take the mean
        # as it would divide by the number of the sequence length, whereas the effective
        # length is only of size k, where k is the number of non-zero scalars in the
        # input mask (since we are only interested in those values)
        last_hidden_states_only_target_aggregated = last_hidden_states_only_target.sum(
            dim=1
        )
        # get the sum for each (batch, hidden states) (sum over the sequence length
        # dim)
        denominator_for_mean = text_bert_indices_target_mask_unsqueezed.sum(dim=1)
        # divide for each (batch, hidden states) by the denominator to get the mean
        last_hidden_states_only_target_aggregated = (
            last_hidden_states_only_target_aggregated / denominator_for_mean
        )
        # dropout before dense layer, as in most other tsc models
        last_hidden_states_only_target_aggregated = self.dropout(
            last_hidden_states_only_target_aggregated
        )
        # dense layer
        logits = self.dense(last_hidden_states_only_target_aggregated)

        return logits



# -*- coding: utf-8 -*-
# file: squeeze_embedding.py
# author: songyouwei <youwei0314@gmail.com>
# Copyright (C) 2018. All Rights Reserved.


import torch
import torch.nn as nn
import numpy as np

class SqueezeEmbedding(nn.Module):
    """
    Squeeze sequence embedding length to the longest one in the batch
    """
    def __init__(self, batch_first=True):
        super(SqueezeEmbedding, self).__init__()
        self.batch_first = batch_first

    def forward(self, x, x_len):
        """
        sequence -> sort -> pad and pack -> unpack ->unsort
        :param x: sequence embedding vectors
        :param x_len: numpy/tensor list
        :return:
        """
        """sort"""
        x_sort_idx = torch.sort(-x_len)[1].long()
        x_unsort_idx = torch.sort(x_sort_idx)[1].long()
        x_len = x_len[x_sort_idx]
        x = x[x_sort_idx]
        """pack"""
        x_emb_p = torch.nn.utils.rnn.pack_padded_sequence(x, x_len, batch_first=self.batch_first)
        """unpack: out"""
        out = torch.nn.utils.rnn.pad_packed_sequence(x_emb_p, batch_first=self.batch_first)  # (sequence, lengths)
        out = out[0]  #
        """unsort"""
        out = out[x_unsort_idx]
        return out



# -*- coding: utf-8 -*-
# file: dynamic_rnn.py
# author: songyouwei <youwei0314@gmail.com>
# Copyright (C) 2018. All Rights Reserved.


import torch
import torch.nn as nn
import numpy as np

class DynamicLSTM(nn.Module):
    def __init__(self, input_size, hidden_size, num_layers=1, bias=True, batch_first=True, dropout=0,
                 bidirectional=False, only_use_last_hidden_state=False, rnn_type = 'LSTM'):
        """
        LSTM which can hold variable length sequence, use like TensorFlow's RNN(input, length...).

        :param input_size:The number of expected features in the input x
        :param hidden_size:The number of features in the hidden state h
        :param num_layers:Number of recurrent layers.
        :param bias:If False, then the layer does not use bias weights b_ih and b_hh. Default: True
        :param batch_first:If True, then the input and output tensors are provided as (batch, seq, feature)
        :param dropout:If non-zero, introduces a dropout layer on the outputs of each RNN layer except the last layer
        :param bidirectional:If True, becomes a bidirectional RNN. Default: False
        :param rnn_type: {LSTM, GRU, RNN}
        """
        super(DynamicLSTM, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bias = bias
        self.batch_first = batch_first
        self.dropout = dropout
        self.bidirectional = bidirectional
        self.only_use_last_hidden_state = only_use_last_hidden_state
        self.rnn_type = rnn_type
        
        if self.rnn_type == 'LSTM': 
            self.RNN = nn.LSTM(
                input_size=input_size, hidden_size=hidden_size, num_layers=num_layers,
                bias=bias, batch_first=batch_first, dropout=dropout, bidirectional=bidirectional)  
        elif self.rnn_type == 'GRU':
            self.RNN = nn.GRU(
                input_size=input_size, hidden_size=hidden_size, num_layers=num_layers,
                bias=bias, batch_first=batch_first, dropout=dropout, bidirectional=bidirectional)
        elif self.rnn_type == 'RNN':
            self.RNN = nn.RNN(
                input_size=input_size, hidden_size=hidden_size, num_layers=num_layers,
                bias=bias, batch_first=batch_first, dropout=dropout, bidirectional=bidirectional)
        

    def forward(self, x, x_len):
        """
        sequence -> sort -> pad and pack ->process using RNN -> unpack ->unsort

        :param x: sequence embedding vectors
        :param x_len: numpy/tensor list
        :return:
        """
        """sort"""
        x_sort_idx = torch.sort(-x_len)[1].long()
        x_unsort_idx = torch.sort(x_sort_idx)[1].long()
        x_len = x_len[x_sort_idx]
        x = x[x_sort_idx]
        """pack"""
        x_emb_p = torch.nn.utils.rnn.pack_padded_sequence(x, x_len, batch_first=self.batch_first)
        
        # process using the selected RNN
        if self.rnn_type == 'LSTM': 
            out_pack, (ht, ct) = self.RNN(x_emb_p, None)
        else: 
            out_pack, ht = self.RNN(x_emb_p, None)
            ct = None
        """unsort: h"""
        ht = torch.transpose(ht, 0, 1)[
            x_unsort_idx]  # (num_layers * num_directions, batch, hidden_size) -> (batch, ...)
        ht = torch.transpose(ht, 0, 1)

        if self.only_use_last_hidden_state:
            return ht
        else:
            """unpack: out"""
            out = torch.nn.utils.rnn.pad_packed_sequence(out_pack, batch_first=self.batch_first)  # (sequence, lengths)
            out = out[0]  #
            out = out[x_unsort_idx]
            """unsort: out c"""
            if self.rnn_type =='LSTM':
                ct = torch.transpose(ct, 0, 1)[
                    x_unsort_idx]  # (num_layers * num_directions, batch, hidden_size) -> (batch, ...)
                ct = torch.transpose(ct, 0, 1)

            return out, (ht, ct)



from argparse import Namespace
from typing import Dict

import torch.nn as nn

from NewsSentiment.SentimentClasses import SentimentClasses
from NewsSentiment.consts import *
from NewsSentiment.dataset import FXDataset
from NewsSentiment.models.FXBaseModel import FXBaseModel


class SeqTwoSeqWithoutTargetMask(FXBaseModel):
    """
    Outputs the class probabilities for each token. So, the output will be:
    (batch, seqlen (150), classnum (3))
    """

    @staticmethod
    def get_language_models():
        return (BERT_BASE_UNCASED,)

    @staticmethod
    def get_input_field_ids():
        return [
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS),
            (BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS_TARGET_MASK),
        ]

    def __init__(self, transformer_models: Dict, opt: Namespace):
        super(SeqTwoSeqWithoutTargetMask, self).__init__()
        self.language_model = transformer_models[BERT_BASE_UNCASED]
        self.dropout = nn.Dropout(opt.dropout)
        # https://pytorch.org/docs/stable/nn.html#linear supports multi-dimensional
        # input; only the last dimension has to be specified for Linear creation

        self.attentionlike_dense = nn.Linear(
            self.language_model.config.hidden_size,
            SentimentClasses.get_num_classes(),
        )

    def forward(self, inputs):
        # get inputs
        text_bert_indices = FXDataset.get_input_by_params(
            inputs, BERT_BASE_UNCASED, FIELD_TEXT_IDS_WITH_SPECIAL_TOKENS
        )

        # prepare inputs
        # for text only, we do not need target specific information, i.e., all text
        # vectors are identical. also, bert, can
        # only process one sequence of size max_seq_len (more specifically, a tensor
        # of size batch_size x max_seq_ken). thus, we select only the first element from
        # the second dimension (the dimensions are: batch, targets, hidden_states)
        text_bert_indices = text_bert_indices[:, 0, :]
        # apply bert
        last_hidden_states, pooler_output, all_hidden_states = self.language_model(
            input_ids=text_bert_indices
        )
        # shape: batch, seqlen, bertdim
        last_hidden_states = self.dropout(last_hidden_states)

        sequence_logits = self.attentionlike_dense(last_hidden_states)

        return sequence_logits



import os
from copy import copy

import torch.nn as nn

from abc import abstractmethod

from transformers import XLNetModel, AlbertModel, BertModel, RobertaModel
from functools import wraps

from NewsSentiment.download import Download


class FXBaseModel(nn.Module):
    def __init__(self):
        super().__init__()

    @staticmethod
    @abstractmethod
    def get_language_models():
        return

    @staticmethod
    @abstractmethod
    def get_input_field_ids():
        return

    def invoke_language_model(self, lm, input_ids, token_type_ids=None):
        type_lm = type(lm)
        if type_lm == XLNetModel:
            last_hidden_state, mems, all_hidden_states = lm(
                input_ids=input_ids,
                token_type_ids=token_type_ids,
            )
        elif type_lm in [AlbertModel, BertModel, RobertaModel]:
            if token_type_ids is None:
                last_hidden_state, pooler_output, hidden_states = lm(
                    input_ids=input_ids,
                )
            else:
                last_hidden_state, pooler_output, hidden_states = lm(
                    input_ids=input_ids,
                    token_type_ids=token_type_ids
                )
        else:
            raise NotImplementedError

        return last_hidden_state


def provide_pretrained(version, pretrained_url):
    """
    Usage:

        @provide_pretrained("v1.0.0", "https://example.com/link/to/state_dict")
        class Example(nn.Module):
            pass
    """
    def decorator(model_class):
        # The actual decorator to use before the class
        wraps(model_class)

        wrapper = __get_pretrained_wrapper_class(model_class)
        wrapper._provide_pretrained_versions[version] = pretrained_url

        return wrapper
    return decorator


def default_pretrained(version):
    """
    Set the version which should be used as the default version and will be used when running with --pretrained.

    Usage:

        @default_pretrained("v1.0.0")
        @provide_pretrained("v1.0.0", "https://example.com/link/to/state_dict")
        class Example(nn.Module):
            pass
    """
    def decorator(model_class):
        # The actual decorator to use before the class
        wraps(model_class)

        wrapper = __get_pretrained_wrapper_class(model_class)
        wrapper._provide_pretrained_default = version

        return wrapper
    return decorator


def model_includes_pretrained(model):
    """
    Checks if a model-class includes the methods to load pretrained models.

    Arguments:
        model Model-class to check.

    Returns:
        True if it includes the functionality.
    """
    return hasattr(model, 'has_pretrained_state_dict') and hasattr(model, 'get_pretrained_state_dict')


__pretrained_wrapper_classes = set()


def __get_pretrained_wrapper_class(base_class):
    if base_class in __pretrained_wrapper_classes:
        return base_class

    class PretrainedWrapper(base_class):
        _provide_pretrained_default = None
        _provide_pretrained_versions = {}

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        @classmethod
        def has_pretrained_state_dict(cls, version=None):
            version = version or cls._provide_pretrained_default
            return version in cls._provide_pretrained_versions

        @classmethod
        def get_pretrained_state_dict(cls, version=None, download_if_not_exists=True, **kwargs):
            path = Download.model_path(cls, version)
            if os.path.isfile(path):
                if download_if_not_exists:
                    Download.download(cls, version, False)
                else:
                    raise FileNotFoundError('State dict not found')
            return cls.load_state_dict(path, **kwargs)

        @classmethod
        def get_pretrained_versions(cls):
            return copy(cls._provide_pretrained_versions)

        @classmethod
        def get_pretrained_source(cls, version=None):
            return cls._provide_pretrained_versions[version or cls._provide_pretrained_default]

        @classmethod
        def get_pretrained_default_version(cls):
            return cls._provide_pretrained_default

    __pretrained_wrapper_classes.add(PretrainedWrapper)
    return PretrainedWrapper



import torch

from NewsSentiment.diskdict import DiskDict
from NewsSentiment.knowledge.knowledgeutils import find_key_original_or_lc
from NewsSentiment.knowledge.mpqasubjectivity.converter import POLARITY2INDEX, PATH_DICT_MPQA_SUBJECTIVITY

__ddict_emolex = DiskDict(PATH_DICT_MPQA_SUBJECTIVITY)
__ddict_emolex_keys_lower = {k.lower(): v for k, v in __ddict_emolex.items()}
__num_emotions = 3
assert len(__ddict_emolex) == 6886, len(__ddict_emolex)
assert len(POLARITY2INDEX) == __num_emotions


def get_num_mpqa_subjectivity_polarities():
    return __num_emotions


def get_mpqa_subjectivity_polarities_as_tensor(term: str):
    emotions = find_key_original_or_lc(__ddict_emolex, __ddict_emolex_keys_lower, term)
    tensor_emotions = torch.zeros(__num_emotions, dtype=torch.long)
    for emotion_index in emotions:
        tensor_emotions[emotion_index] = 1
    return tensor_emotions



from pathlib import Path
import pickle
from typing import Union

from NewsSentiment.fxlogger import get_logger

logger = get_logger()


class DiskDict(dict):
    def __init__(self, sync_path: Union[Path, str]):
        if type(sync_path) == str:
            sync_path = Path(sync_path)
        self.path = sync_path

        if self.path.exists():
            with open(self.path, "rb") as file:
                tmp_dct = pickle.load(file)
                super().update(tmp_dct)
                logger.info(
                    "loaded DiskDict with %s items from %s", len(tmp_dct), self.path
                )

    def sync_to_disk(self):
        with open(self.path, "wb") as file:
            tmp_dct = super().copy()
            pickle.dump(tmp_dct, file)
            # print(f"saved DiskDict with {len(tmp_dct)} items to {self.path}")
